# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import inspect
import logging
import os
from typing import (
    Any, Callable, cast, Dict, final, Iterable, Iterator, List, Optional,
    overload, Sequence, Tuple, Type, TYPE_CHECKING, Union)

import numpy
from numpy import floating
from numpy.typing import NDArray
from typing_extensions import Never, TypeAlias

import neo
from neo.io.baseio import BaseIO  # type: ignore[import]
from pyNN.descriptions import TemplateEngine
from pyNN import descriptions
from pyNN.random import NumpyRNG
from pyNN.space import BaseStructure

from spinn_utilities.log import FormatAdapter
from spinn_utilities.logger_utils import warn_once
from spinn_utilities.overrides import overrides

from spinn_front_end_common.utilities.exceptions import ConfigurationException

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.exceptions import SpynnakerException
from spynnaker.pyNN.models.abstract_models import SupportsStructure
from spynnaker.pyNN.models.abstract_pynn_model import AbstractPyNNModel
from spynnaker.pyNN.models.common import PopulationApplicationVertex
from spynnaker.pyNN.models.recorder import Recorder
from spynnaker.pyNN.types import IoDest
from spynnaker.pyNN.utilities.neo_buffer_database import NeoBufferDatabase
from spynnaker.pyNN.types import Selector
from spynnaker.pyNN.utilities.utility_calls import get_neo_io

from .population_base import PopulationBase
from .population_view import PopulationView, IDMixin

if TYPE_CHECKING:
    from pyNN.neuron.standardmodels.electrodes import NeuronCurrentSource
    from spynnaker.pyNN.models.common.types import Names, Values
    from spynnaker.pyNN.models.common.parameter_holder import ParameterHolder
    __Values: TypeAlias = Values  # Stupid PyDev!

logger = FormatAdapter(logging.getLogger(__file__))
_CellType: TypeAlias = Union[AbstractPyNNModel, PopulationApplicationVertex]
_CellTypeArg: TypeAlias = Union[Type[AbstractPyNNModel], _CellType]
_ParamDict: TypeAlias = Dict[str, Any]


class Population(PopulationBase):
    """
    PyNN population object.
    """
    # pylint: disable=redefined-builtin
    __slots__ = (
        "__annotations",
        "__celltype",
        "__first_id",
        "__last_id",
        "__positions",
        "__recorder",
        "__size",
        "__structure",
        "__vertex")

    def __init__(
            self, size: Union[int, float, None], cellclass: _CellTypeArg,
            cellparams: Optional[_ParamDict] = None,
            structure: Optional[BaseStructure] = None,
            initial_values: Optional[Dict[str, float]] = None,
            label: Optional[str] = None,
            additional_parameters: Optional[_ParamDict] = None,
            **additional_kwargs: _ParamDict):
        """
        :param size: The number of neurons in the population
        :param cellclass: The implementation of the individual neurons.
        :param cellparams: Parameters to pass to ``cellclass`` if it
            is a class to instantiate. Must be ``None`` if ``cellclass`` is an
            instantiated object.
        :param structure:
        :param initial_values: Initial values of state variables
        :param label: A label for the population
        :param additional_parameters:
            Additional parameters to pass to the vertex creation function.
        :param additional_kwargs:
            A nicer way of allowing additional things
        """
        # Deal with the kwargs!
        additional: _ParamDict = dict()
        if additional_parameters is not None:
            additional.update(additional_parameters)
        if additional_kwargs:
            additional.update(additional_kwargs)

        # build our initial objects
        self.__celltype: AbstractPyNNModel
        self.__vertex: PopulationApplicationVertex
        model = self.__create_model(cellclass, cellparams)
        realsize = self.__roundsize(size, label)
        self.__create_vertex(model, realsize, label, additional)
        self.__recorder = Recorder(population=self, vertex=self.__vertex)

        # structure should be a valid Space.py structure type.
        # generation of positions is deferred until needed.
        self.__structure = structure
        self.__positions: Optional[numpy.ndarray] = None
        if isinstance(self.__vertex, SupportsStructure):
            self.__vertex.set_structure(structure)

        # add objects to the SpiNNaker control class
        SpynnakerDataView.add_vertex(self.__vertex)

        # initialise common stuff
        if realsize is None:
            realsize = self.__vertex.n_atoms
        self.__size = realsize
        self.__annotations: Dict[str, Any] = dict()

        # things for pynn demands
        self.__first_id, self.__last_id = SpynnakerDataView.add_population(
            self)

        # set up initial values if given
        if initial_values:
            for variable, value in initial_values.items():
                self.__vertex.set_initial_state_values(variable, value)

    def __iter__(self) -> Iterator[PopulationView]:
        """
        Iterate over local cells.
        """
        for _id in range(self.__size):
            yield IDMixin(self, _id)

    def __getitem__(self, index_or_slice: Selector) -> PopulationView:
        if isinstance(index_or_slice, int):
            return IDMixin(self, index_or_slice)
        else:
            return PopulationView(self, index_or_slice)

    def all(self) -> Iterator[PopulationView]:
        """
        :returns: Iterator over cell IDs on all MPI nodes.
        """
        for _id in range(self.__size):
            yield IDMixin(self, _id)

    @property
    def annotations(self) -> Dict[str, Any]:
        """
        The annotations given by the end user.
        """
        return self.__annotations

    @property
    def celltype(self) -> AbstractPyNNModel:
        """
        The Model uses for this population.
        """
        return self.__celltype

    def can_record(self, variable: str) -> bool:
        """
        Determine whether `variable` can be recorded from this population.

        :param variable: The variable to answer the question about
        :returns: True if the variable can be recorded, False otherwise.
        """
        return variable in self.__vertex.get_recordable_variables()

    @overrides(PopulationBase.record)
    def record(self, variables: Names, to_file: IoDest = None,
               sampling_interval: Optional[int] = None) -> None:
        self.__recorder.record(
            variables, to_file, sampling_interval, indexes=None)

    def sample(self, n: int, rng: Optional[NumpyRNG] = None) -> PopulationView:
        """
        Randomly sample `n` cells from the Population, and return a
        PopulationView object.

        :param n: The number of cells to put in the view.
        :param rng: The random number generator to use
        :returns: A PopulationView over n random cells.
        """
        if not rng:
            rng = NumpyRNG()
        indices = rng.permutation(
            numpy.arange(len(self), dtype=numpy.int32))[0:n]
        return PopulationView(
            self, indices,
            label=f"Random sample size {n} from {self.label}")

    @overrides(PopulationBase.write_data)
    def write_data(self, io: Union[str, BaseIO], variables: Names = 'all',
                   gather: bool = True, clear: bool = False,
                   annotations: Optional[Dict[str, Any]] = None) -> None:
        self._check_params(gather, annotations)

        if isinstance(io, str):
            extension = os.path.splitext(io)[1][1:]
            if extension == "csv":
                self.__recorder.write_data(
                    io, variables, annotations=annotations)
                return
            io = get_neo_io(io)

        data = self.__recorder.extract_neo_block(
            variables, None, clear, annotations)
        # write the neo block to the file
        io.write(data)

    def describe(self, template: str = 'population_default.txt',
                 engine:  Optional[Union[str, TemplateEngine]] = 'default'
                 ) -> Union[str, Dict[str, Any]]:
        """
        Returns a human-readable description of the population.

        The output may be customised by specifying a different template
        together with an associated template engine (see
        :mod:`pyNN.descriptions`).

        If ``template`` is ``None``, then a dictionary containing the template
        context will be returned.

        :param template: Template filename
        :param engine: Template substitution engine
        :returns: Human-readable description as a string or dict
        """
        context: Dict[str, Any] = {
            "label": self.label,
            "celltype": self.celltype.describe(template=None),
            "structure": None,
            "size": self.size,
            "size_local": self.size,
            "first_id": self.first_id,
            "last_id": self.last_id,
        }
        context.update(self.annotations)
        if self.size > 0:
            parameters = self.__vertex.get_parameters()
            cell_parameters: Union[str, ParameterHolder]
            if parameters:
                cell_parameters = self.__vertex.get_parameter_values(
                    parameters, 0)
            else:
                cell_parameters = "No cell parameters"
            context.update({
                "local_first_id": 0,
                "cell_parameters": cell_parameters
            })
        if self.structure:
            context["structure"] = self.structure.describe(template=None)
        return descriptions.render(engine, template, context)

    def _end(self) -> None:
        """
        Do final steps at the end of the simulation.
        """
        for variable in self.__recorder.write_to_files_indicators:
            if self.__recorder.write_to_files_indicators[variable]:
                self.write_data(
                    io=self.__recorder.write_to_files_indicators[variable],
                    variables=[variable])

    @overrides(PopulationBase.get_data)
    def get_data(
            self, variables: Names = 'all',
            gather: bool = True, clear: bool = False, *,
            annotations: Optional[Dict[str, Any]] = None) -> neo.Block:
        self._check_params(gather, annotations)
        return self.__recorder.extract_neo_block(
            variables, None, clear, annotations)

    def spinnaker_get_data(
            self, variable: str, as_matrix: bool = False,
            view_indexes: Optional[Sequence[int]] = None) -> NDArray[floating]:
        """
        SsPyNNaker specific method for getting data as a numpy array,
        instead of the Neo-based object

        :param variable: a single variable name.
        :param as_matrix: If set True the data is returned as a 2d matrix
        :param view_indexes: The indexes for which data should be returned.
            If ``None``, all data (view_index = data_indexes)
        :return: array of the data
        """
        warn_once(
            logger, "spinnaker_get_data is non-standard PyNN and therefore "
            "will not be portable to other simulators.")
        with NeoBufferDatabase() as db:
            return db.spinnaker_get_data(self.__recorder.recording_label,
                                         variable, as_matrix, view_indexes)

    @overrides(PopulationBase.get_spike_counts)
    def get_spike_counts(self, gather: bool = True) -> Dict[int, int]:
        self._check_params(gather)
        with NeoBufferDatabase() as db:
            return db.get_spike_counts(self.__recorder.recording_label)

    def find_units(self, variable: str) -> str:
        """
        Get the units of a variable.

        :param variable: The name of the variable
        :return: The units of the variable
        """
        return self.__vertex.get_units(variable)

    def set(self, **parameters: Values) -> None:
        """
        Set one or more parameters for every cell in the population.
        For example::

            p.set(tau_m=20.0).

        :param parameters:
            The parameters to set and the values to set them to.
            The type of each parameter depends on the parameter; it's often a
            float, but not always.
        :raises SimulatorRunningException: If `sim.run` is currently running
        :raises SimulatorNotSetupException: If called before `sim.setup`
        :raises SimulatorShutdownException: If called after `sim.end`
        """
        SpynnakerDataView.check_user_can_act()
        for parameter, value in parameters.items():
            self.__vertex.set_parameter_values(parameter, value)

    def initialize(self, **kwargs: Values) -> None:
        """
        Set initial values of state variables, e.g. the membrane potential.
        Values passed to ``initialize()`` may be:

        * single numeric values (all neurons set to the same value), or
        * :py:class:`~spynnaker.pyNN.RandomDistribution` objects, or
        * lists / arrays of numbers of the same size as the population
          mapping functions, where a mapping function accepts a single
          argument (the cell index) and returns a single number.

        Values should be expressed in the standard PyNN units (i.e.
        millivolts, nanoamps, milliseconds, microsiemens, nanofarads,
        event per second).

        Examples::

            p.initialize(v=-70.0)
            p.initialize(v=rand_distr, gsyn_exc=0.0)
            p.initialize(v=lambda i: -65 + i / 10.0)
        """
        SpynnakerDataView.check_user_can_act()
        if SpynnakerDataView.is_ran_last():
            warn_once(
                logger, "Calling initialize without reset will have no effect."
                " If you want to set the current value of a state variable"
                " consider calling the non-PyNN function set_state instead.")
        for variable, value in kwargs.items():
            self.__vertex.set_initial_state_values(variable, value)

    @property
    def initial_values(self) -> ParameterHolder:
        """
        The initial values of the state variables.

        .. note::
            These values will be the same as the values set with the last call
            to initialize rather than the actual initial values if this call
            has been made.
        """
        SpynnakerDataView.check_user_can_act()
        return self.__vertex.get_initial_state_values(
            self.__vertex.get_state_variables())

    def set_state(self, **kwargs: Values) -> None:
        """
        Set current values of state variables, e.g. the membrane potential.
        Values passed to ``set_state()`` may be:

        * single numeric values (all neurons set to the same value), or
        * :py:class:`~spynnaker.pyNN.RandomDistribution` objects, or
        * lists / arrays of numbers of the same size as the population
          mapping functions, where a mapping function accepts a single
          argument (the cell index) and returns a single number.

        Values should be expressed in the standard PyNN units (i.e.
        millivolts, nanoamps, milliseconds, microsiemens, nanofarads,
        event per second).

        Examples::

            p.set_state(v=-70.0)
            p.set_state(v=rand_distr, gsyn_exc=0.0)
            p.set_state(v=lambda i: -65 + i / 10.0)
        """
        SpynnakerDataView.check_user_can_act()
        warn_once(
            logger, "set_state is non-standard PyNN and therefore "
            "will not be portable to other simulators.")
        for variable, value in kwargs.items():
            self.__vertex.set_current_state_values(variable, value)

    @property
    def current_values(self) -> ParameterHolder:
        """
        Get the current values of the state variables.
        """
        SpynnakerDataView.check_user_can_act()
        warn_once(
            logger, "current_values is non-standard PyNN and therefore "
            "will not be portable to other simulators.")
        return self.__vertex.get_current_state_values(
            self.__vertex.get_state_variables())

    @property
    @overrides(PopulationBase.positions)
    def positions(self) -> NDArray[floating]:
        if self.__positions is None:
            if self.__structure is None:
                raise ValueError("attempted to retrieve positions "
                                 "for an unstructured population")
            self.__positions = self.__structure.generate_positions(
                self.__vertex.n_atoms)
        return self.__positions.T

    @positions.setter
    def positions(self, positions: NDArray[numpy.floating]) -> None:
        """
        Sets all the positions in the population.
        """
        self.__positions = positions

    @property
    @overrides(PopulationBase.all_cells)
    def all_cells(self) -> List[IDMixin]:
        return [IDMixin(self, _id) for _id in range(self.__size)]

    @property
    @overrides(PopulationBase.position_generator)
    def position_generator(self) -> Callable[[int], NDArray[numpy.floating]]:
        return lambda i: self.positions[:, i]

    @property
    def first_id(self) -> int:
        """
        The ID of the first member of the population.
        """
        return self.__first_id

    @property
    def last_id(self) -> int:
        """
        The ID of the last member of the population.
        """
        return self.__last_id

    @property
    @overrides(PopulationBase._vertex)
    def _vertex(self) -> PopulationApplicationVertex:
        return self.__vertex

    @property
    def _recorder(self) -> Recorder:
        return self.__recorder

    @property
    @overrides(PopulationBase._view_range)
    def _view_range(self) -> Tuple[int, int]:
        return 0, self.size - 1

    @property
    def conductance_based(self) -> bool:
        """
        Whether the population uses conductance inputs
        """
        return self.__vertex.conductance_based

    def get(self, parameter_names: Names,
            gather: bool = True, simplify: bool = True) -> ParameterHolder:
        """
        Get the values of a parameter for every local cell in the population.

        :param parameter_names: Name of parameter. This is either a single
            string or a list of strings
        :param gather: Ignored. Purely for PyNN compatibility
        :param simplify: Ignored. Purely for PyNN compatibility
        :return: A single list of values (or possibly a single value) if
            paramter_names is a string, or a dict of these if parameter names
            is a list.
        """
        self._check_params(gather)
        if simplify is not True:
            warn_once(
                logger, "The simplify value is ignored if not set to true")

        return self.__vertex.get_parameter_values(parameter_names)

    @overload
    def id_to_index(self, id: int) -> int:  # @ReservedAssignment
        ...

    @overload
    def id_to_index(
            self, id: Iterable[int]) -> List[int]:  # @ReservedAssignment
        ...

    def id_to_index(self, id: Union[int, Iterable[int]]
                    ) -> Union[int, List[int]]:  # @ReservedAssignment
        """
        Given the ID(s) of cell(s) in the Population, return its (their)
        index (order in the Population).

        Defined by
        https://neuralensemble.org/docs/PyNN/reference/populations.html

        :param id:
        :returns: Index of cell(s) in the Population
        """
        # pylint: disable=redefined-builtin
        if not numpy.iterable(id):
            id = cast(int, id)
            if not self.__first_id <= id <= self.__last_id:
                raise ValueError(
                    f"id should be in the range [{self.__first_id},"
                    f"{self.__last_id}], actually {id}")
            return int(id - self.__first_id)  # assume IDs are consecutive
        return [_id - self.__first_id for _id in id]

    @overload
    def index_to_id(self, index: int) -> int:
        ...

    @overload
    def index_to_id(self, index: Iterable[int]) -> List[int]:
        ...

    def index_to_id(self, index: Union[int, Iterable[int]]
                    ) -> Union[int, List[int]]:
        """
        Given the index (order in the Population) of cell(s) in the
        Population, return their ID(s)

        :param index:
        :returns: The ID(s) of the cell(s) in the Population
        """
        if not numpy.iterable(index):
            index = cast(int, index)
            if index > self.__last_id - self.__first_id:
                raise ValueError(
                    "indexes should be in the range [0,"
                    f"{self.__last_id - self.__first_id}], actually {index}")
            return int(index + self.__first_id)
        # this assumes IDs are consecutive
        return [_index + self.__first_id for _index in index]

    def id_to_local_index(self, cell_id: Union[int, Iterable[int]]) -> Never:
        """
        Given the ID(s) of cell(s) in the Population, return its (their)
        index (order in the Population), counting only cells on the local
        MPI node.

        Defined by
        https://neuralensemble.org/docs/PyNN/reference/populations.html

        :param cell_id:
        """
        raise NotImplementedError(
            "sPyNNaker does not currently do id_to_local_index")

    @overrides(PopulationBase.inject)
    def inject(self, current_source: NeuronCurrentSource) -> None:
        """
        Connect a current source to all cells in the Population.

        Defined by
        https://neuralensemble.org/docs/PyNN/reference/populations.html
        """
        # Pass this into the vertex
        self.__vertex.inject(current_source, [n for n in range(self.__size)])

    def __len__(self) -> int:
        """
        Get the total number of cells in the population.
        """
        return self.__size

    @property
    def label(self) -> str:
        """
        The label of the population.
        """
        return self.__vertex.label or ""  # Should never be empty

    @label.setter
    def label(self, label: str) -> Never:
        raise NotImplementedError(
            "As label is used as an ID it can not be changed")

    @property
    @overrides(PopulationBase.structure)
    def structure(self) -> Optional[BaseStructure]:
        """
        The structure for the population.
        """
        return self.__structure

    # NON-PYNN API CALL
    def add_placement_constraint(
            self, x: int, y: int, p: Optional[int] = None) -> None:
        """
        Add a placement constraint.

        :param x: The x-coordinate of the placement constraint
        :param y: The y-coordinate of the placement constraint
        :param p: The processor ID of the placement constraint (optional)
        :raises SimulatorRunningException: If `sim.run` is currently running
        :raises SimulatorNotSetupException: If called before `sim.setup`
        :raises SimulatorShutdownException: If called after `sim.end`
        """
        self.__vertex.set_fixed_location(x, y, p)

    # NON-PYNN API CALL
    def set_max_atoms_per_core(
            self, max_atoms_per_core: Union[int, Tuple[int, ...]]) -> None:
        """
        Supports the setting of this population's max atoms per
        dimension per core.

        :param max_atoms_per_core:
            the new value for the max atoms per dimension per core.
        :raises SimulatorRunningException: If `sim.run` is currently running
        :raises SimulatorNotSetupException: If called before `sim.setup`
        :raises SimulatorShutdownException: If called after `sim.end`
        """
        SpynnakerDataView.check_user_can_act()
        ct = self.celltype
        if isinstance(ct, AbstractPyNNModel):
            cap = ct.absolute_max_atoms_per_core
            if numpy.prod(max_atoms_per_core) > cap:
                raise SpynnakerException(
                    f"Set the max_atoms_per_core to {max_atoms_per_core} "
                    f"blocked as the current limit for the model is {cap}")
        self.__vertex.set_max_atoms_per_dimension_per_core(max_atoms_per_core)

    @property
    def size(self) -> int:
        """
        The number of neurons in the population.
        """
        return self.__vertex.n_atoms

    @staticmethod
    def __create_model(
            cell_class: _CellTypeArg,
            cell_params: Optional[_ParamDict]) -> _CellType:
        """
        :param cell_class: The implementation of the individual neurons.
        :param cell_params: Parameters to pass to ``cell_class`` if it
            is a class to instantiate. Must be ``None`` if ``cell_class`` is an
            instantiated object.
        """
        if inspect.isclass(cell_class):
            if cell_params is None:
                model = cell_class()
            else:
                model = cell_class(**cell_params)
            assert isinstance(model, (
                AbstractPyNNModel, PopulationApplicationVertex))
        elif cell_params:
            raise ConfigurationException(
                "cell_class is an instance which includes params so "
                "cell_params must be None")
        else:
            model = cell_class
        return model

    def __create_vertex_from_model(
            self, model: AbstractPyNNModel, size: Optional[int],
            label: Optional[str], additional_parameters: _ParamDict) -> None:
        """
        Worker for :meth:`__create_vertex` to handle the case where we really
        have a model.
        """
        self.__celltype = model
        # pylint: disable=protected-access
        if size is not None and size <= 0:
            raise ConfigurationException(
                "A population cannot have a negative or zero size.")
        parameters = model._get_default_population_parameters()
        if additional_parameters:
            # check that the additions are suitable. report wrong ones
            # and ignore
            parameters = self.__process_additional_params(
                additional_parameters, parameters)
        self.__vertex = model.create_vertex(
            size or 1, label or f"{model.name} vertex", **parameters)

    def __init_with_supplied_vertex(
            self, model: PopulationApplicationVertex, size: Optional[int],
            label: Optional[str], additional_parameters: _ParamDict) -> None:
        """
        Worker for :meth:`__create_vertex` to handle the case where we have a
        user-supplied vertex.
        """
        if additional_parameters:
            raise ConfigurationException(
                "Cannot accept additional parameters "
                f"{additional_parameters} when the cell is a vertex")
        # Use a synthetic model
        self.__celltype = _VertexHolder(model)
        self.__vertex = model
        if size is not None and size != self.__vertex.n_atoms:
            raise ConfigurationException(
                "Vertex size does not match Population size")
        if label is not None:
            self.__vertex.set_label(label)

    def __create_vertex(
            self, model: _CellType, size: Optional[int], label: Optional[str],
            additional_parameters: _ParamDict) -> None:
        """
        :param model: The implementation of the individual neurons.
        :param size:
        :param label:
        :param additional_parameters:
            Additional parameters to pass to the vertex creation function.
        """
        # Use a provided model to create a vertex
        if isinstance(model, AbstractPyNNModel):
            self.__create_vertex_from_model(
                model, size, label, additional_parameters)

        # Use a provided application vertex directly
        elif isinstance(model, PopulationApplicationVertex):
            self.__init_with_supplied_vertex(
                model, size, label, additional_parameters)

        # Fail on anything else
        else:
            raise ConfigurationException(
                "Model must be either an AbstractPyNNModel or a"
                " PopulationApplicationVertex")

    @staticmethod
    def create(
            cellclass: _CellTypeArg, cellparams: Optional[_ParamDict] = None,
            n: int = 1) -> 'Population':
        """
        Pass through method to the constructor defined by PyNN.
        Create ``n`` cells all of the same type.

        :param cellclass: see :py:meth:`~.Population.__init__`
        :param cellparams: see :py:meth:`~.Population.__init__`
        :param n: see :py:meth:`~.Population.__init__` (``size`` parameter)
        :return: A New Population
        """
        return Population(size=n, cellclass=cellclass, cellparams=cellparams)

    @staticmethod
    def __process_additional_params(
            additional_parameters: _ParamDict,
            population_parameters: _ParamDict) -> _ParamDict:
        """
        Essential method for allowing things like splitter objects at
        population level.

        :param additional_parameters:
            the additional parameters handed down from user
        :param population_parameters:
            the additional parameters the vertex can support.
        :return: the list of parameters that are accepted.
        """
        for key in additional_parameters.keys():
            if key in population_parameters:
                population_parameters[key] = additional_parameters[key]
            else:
                logger.warning("additional_parameter {} will be ignored", key)
        return population_parameters

    @staticmethod
    def __roundsize(
            size: Union[int, float, None],
            label: Optional[str]) -> Optional[int]:
        # External device population can have a size of None so accept for now
        if size is None or isinstance(size, int):
            return size
        # Allow a float which has a near int value
        temp = int(round(size))
        if abs(temp - size) < 0.001:
            logger.warning("Size of the population {} rounded "
                           "from {} to {}. Please use int values for size",
                           label, size, temp)
            return temp
        raise ConfigurationException(
            f"Size of a population with label {label} must be an int,"
            f" received {size}")


@final
class _VertexHolder(AbstractPyNNModel):
    """
    A simplistic model that just holds its supplied vertex.
    It has nothing to configure.
    """
    __slots__ = ("__vertex", )
    default_population_parameters = {}

    def __init__(self, vertex: PopulationApplicationVertex):
        """
        :param vertex: Vertex to hold and return
        """
        self.__vertex = vertex

    @property
    def vertex(self) -> PopulationApplicationVertex:
        """
        The vertex passed into the init.
        """
        return self.__vertex

    @overrides(AbstractPyNNModel.create_vertex)
    def create_vertex(
            self, n_neurons: int, label: str) -> PopulationApplicationVertex:
        # The parameters are ignored; the vertex already exists
        return self.__vertex




# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import math
from typing import Optional, Sequence, TYPE_CHECKING

from numpy import uint32, integer
from numpy.typing import NDArray
import numpy.random

from pyNN.random import NumpyRNG

from spinn_utilities.overrides import overrides

from pacman.model.graphs.common import Slice

from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.utilities import utility_calls
from spynnaker.pyNN.exceptions import SpynnakerException

from .abstract_connector import AbstractConnector
from .abstract_generate_connector_on_machine import (
    AbstractGenerateConnectorOnMachine, ConnectorIDs)
from .abstract_generate_connector_on_host import (
    AbstractGenerateConnectorOnHost)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)


class MultapseConnector(AbstractGenerateConnectorOnMachine,
                        AbstractGenerateConnectorOnHost):
    """
    Create a multapse connector. The size of the source and destination
    populations are obtained when the projection is connected. The number
    of synapses is specified. when instantiated, the required number of
    synapses is created by selecting at random from the source and target
    populations with replacement. Uniform selection probability is assumed.
    """
    __slots__ = (
        "__allow_self_connections",
        "__num_synapses",
        "__post_slices",
        "__synapses_per_edge",
        "__with_replacement",
        "__rng")

    def __init__(self, n: int, allow_self_connections: bool = True,
                 with_replacement: bool = True, rng: Optional[NumpyRNG] = None,
                 safe: bool = True, verbose: bool = False,
                 callback: None = None):
        """
        :param n:
            This is the total number of synapses in the connection.
        :param allow_self_connections:
            Allow a neuron to connect to itself or not.
        :param with_replacement:
            When selecting, allow a neuron to be re-selected or not.
        :param safe:
            Whether to check that weights and delays have valid values.
            If ``False``, this check is skipped.
        :param verbose:
            Whether to output extra information about the connectivity to a
            CSV file
        :param rng:
            Seeded random number generator, or ``None`` to make one when
            needed.
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.
        """
        super().__init__(safe, callback, verbose)
        # We absolutely require an integer at this point!
        self.__num_synapses = self._roundsize(n, "MultapseConnector")
        self.__allow_self_connections = allow_self_connections
        self.__with_replacement = with_replacement
        self.__post_slices: Sequence[Slice] = ()
        self.__synapses_per_edge: Optional[NDArray[integer]] = None
        self.__rng = rng

    def set_projection_information(
            self, synapse_info: SynapseInformation) -> None:
        super().set_projection_information(synapse_info)
        n_pairs = synapse_info.n_post_neurons * synapse_info.n_pre_neurons
        if not self.__with_replacement and self.__num_synapses > n_pairs:
            raise SpynnakerException(
                "FixedTotalNumberConnector will not work when "
                "with_replacement=False and n > n_pre * n_post")
        if (not self.__with_replacement and
                not self.__allow_self_connections and
                self.__num_synapses == n_pairs and
                synapse_info.pre_population == synapse_info.post_population):
            raise SpynnakerException(
                "FixedNumberPostConnector will not work when "
                "with_replacement=False, allow_self_connections=False "
                "and n = n_pre * n_post")

    def get_rng_next(self, num_synapses: int, prob_connect: Sequence[float],
                     rng: NumpyRNG) -> NDArray[integer]:
        """
        Get the required RNGs.

        :param num_synapses:
            The number of synapses to make random numbers for in this call
        :param prob_connect: The probability of connection
        :param rng: The random number generator
        :returns: The multinomial distribution
        """
        # Below is how numpy does multinomial internally...
        size = len(prob_connect)
        multinomial = numpy.zeros(size, int)
        total = 1.0
        dn = num_synapses
        for j in range(0, size - 1):
            multinomial[j] = rng.next(
                1, distribution="binomial",
                parameters={'n': dn, 'p': prob_connect[j] / total})
            dn = dn - multinomial[j]
            if dn <= 0:
                break
            total = total - prob_connect[j]
        if dn > 0:
            multinomial[size - 1] = dn

        return multinomial

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_maximum(
            synapse_info.delays, self.__num_synapses, synapse_info)

    @overrides(AbstractConnector.get_delay_minimum)
    def get_delay_minimum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_minimum(
            synapse_info.delays, self.__num_synapses, synapse_info)

    def _update_synapses_per_post_vertex(
            self, post_slices: Sequence[Slice], n_pre_atoms: int,
            rng: NumpyRNG) -> None:
        if (self.__synapses_per_edge is None or
                len(self.__post_slices) != len(post_slices)):
            n_post_atoms = sum(
                post.n_atoms for post in post_slices if post is not None)
            n_connections = n_pre_atoms * n_post_atoms
            if (not self.__with_replacement and
                    n_connections < self.__num_synapses):
                raise SpynnakerException(
                    "FixedNumberTotalConnector will not work correctly when "
                    "with_replacement=False & num_synapses > n_pre * n_post")
            prob_connect = [
                n_pre_atoms * post.n_atoms / float(n_connections)
                if n_pre_atoms and post.n_atoms else 0 for post in post_slices]
            # Use the multinomial directly if possible
            if hasattr(rng, "rng") and hasattr(rng.rng, "multinomial"):
                self.__synapses_per_edge = rng.rng.multinomial(
                    self.__num_synapses, prob_connect)
            else:
                self.__synapses_per_edge = self.get_rng_next(
                    self.__num_synapses, prob_connect, rng)
            assert self.__synapses_per_edge is not None
            if sum(self.__synapses_per_edge) != self.__num_synapses:
                raise SpynnakerException(
                    f"{sum(self.__synapses_per_edge)} of "
                    f"{self.__num_synapses} synapses generated")
            self.__post_slices = post_slices

    def _get_n_connections(self, post_slice_index: int) -> int:
        if not self.__synapses_per_edge:
            return 0
        return self.__synapses_per_edge[post_slice_index]

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:
        # If the chance of there being a connection in the slice is almost 0,
        # there will probably be at least 1 connection somewhere
        prob_in_slice = min(
            n_post_atoms / float(synapse_info.n_post_neurons),
            1.0)
        max_in_slice = max(utility_calls.get_probable_maximum_selected(
            self.__num_synapses, self.__num_synapses, prob_in_slice), 1)

        # Similarly if the chance of there being one in a row is 0, there will
        # probably be 1
        prob_in_row = 1.0 / synapse_info.n_pre_neurons
        n_connections = max(utility_calls.get_probable_maximum_selected(
            self.__num_synapses, max_in_slice, prob_in_row), 1)

        if min_delay is None or max_delay is None:
            return int(math.ceil(n_connections))

        return self._get_n_connections_from_pre_vertex_with_delay_maximum(
            synapse_info.delays,
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            n_connections, min_delay, max_delay, synapse_info)

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        prob_of_choosing_post_atom = 1.0 / synapse_info.n_post_neurons
        return utility_calls.get_probable_maximum_selected(
            self.__num_synapses, self.__num_synapses,
            prob_of_choosing_post_atom)

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(self, synapse_info: SynapseInformation) -> float:
        return self._get_weight_maximum(
            synapse_info.weights, self.__num_synapses, synapse_info)

    @overrides(AbstractGenerateConnectorOnHost.create_synaptic_block)
    def create_synaptic_block(
            self, post_slices: Sequence[Slice], post_vertex_slice: Slice,
            synapse_type: int, synapse_info: SynapseInformation) -> NDArray:
        # update the synapses as required, and get the number of connections
        rng = self.__rng or NumpyRNG()
        post_slice_index = post_slices.index(post_vertex_slice)
        self._update_synapses_per_post_vertex(
            post_slices, synapse_info.n_pre_neurons, rng)
        n_connections = self._get_n_connections(post_slice_index)
        if n_connections == 0:
            return numpy.zeros(0, dtype=self.NUMPY_SYNAPSES_DTYPE)

        # set up array for synaptic block
        block = numpy.zeros(
            n_connections, dtype=AbstractConnector.NUMPY_SYNAPSES_DTYPE)

        # Create pairs between the pre- and post-vertex slices
        pairs = numpy.mgrid[0:synapse_info.n_pre_neurons,
                            post_vertex_slice.as_slice()].T.reshape((-1, 2))

        # Deal with case where self-connections aren't allowed
        if (not self.__allow_self_connections and
                synapse_info.pre_population is synapse_info.post_population):
            pairs = pairs[pairs[:, 0] != pairs[:, 1]]

        # Now do the actual random choice from the available connections
        try:
            chosen = numpy.random.choice(
                pairs.shape[0], size=n_connections,
                replace=self.__with_replacement)
        except Exception as e:
            raise SpynnakerException(
                "MultapseConnector: The number of connections is too large "
                "for sampling without replacement; "
                "reduce the value specified in the connector") from e

        # Set up synaptic block
        block["source"] = pairs[chosen, 0]
        block["target"] = pairs[chosen, 1]
        block["weight"] = self._generate_weights(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["delay"] = self._generate_delays(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["synapse_type"] = synapse_type
        return block

    def __repr__(self) -> str:
        return f"MultapseConnector({self.__num_synapses})"

    @property
    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_id)
    def gen_connector_id(self) -> int:
        return ConnectorIDs.FIXED_TOTAL_NUMBER_CONNECTOR.value

    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_params)
    def gen_connector_params(
            self, synapse_info: SynapseInformation) -> NDArray[uint32]:
        allow_self = (
            self.__allow_self_connections or
            synapse_info.pre_population != synapse_info.post_population)
        return numpy.array([
            int(allow_self),
            int(self.__with_replacement),
            self.__num_synapses], dtype=uint32)

    @property
    @overrides(
        AbstractGenerateConnectorOnMachine.gen_connector_params_size_in_bytes)
    def gen_connector_params_size_in_bytes(self) -> int:
        return (3 * BYTES_PER_WORD)

    @overrides(AbstractConnector.validate_connection)
    def validate_connection(
            self, application_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        if self.generate_on_machine(synapse_info):
            utility_calls.check_rng(self.__rng, "MultapseConnector")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from dataclasses import dataclass
from typing import (
    Dict, List, NamedTuple, Optional, Sequence, Tuple, TYPE_CHECKING, cast)

import numpy
from numpy import uint32
from numpy.typing import NDArray

from pacman.model.graphs.common import Slice
from pacman.model.placements import Placement
from pacman.model.routing_info import (
    AppVertexRoutingInfo, BaseKeyAndMask)
from pacman.utilities.utility_calls import allocator_bits_needed
from pacman.model.graphs.application import ApplicationVirtualVertex
from spinn_front_end_common.interface.ds import (
    DataType, DataSpecificationBase)

from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.neuron.master_pop_table import (
    MasterPopTableAsBinarySearch)
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    AbstractSynapseDynamicsStructural)
from spynnaker.pyNN.utilities.bit_field_utilities import (
    get_sdram_for_bit_field_region, get_bitfield_key_map_data,
    write_bitfield_init_data, is_sdram_poisson_source)
from spynnaker.pyNN.models.common import PopulationApplicationVertex
from spynnaker.pyNN.models.spike_source import SpikeSourcePoissonVertex
from spynnaker.pyNN.types import WeightScales
from .synaptic_matrix_app import SynapticMatrixApp

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neuron.population_vertex import (
        PopulationVertex)
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)

# 1 for synaptic matrix region
# 1 for master pop region
# 1 for bitfield filter region
# 1 for structural region
# 1 for n_edges
# 1 for post_vertex_slice.lo_atom
# 1 for post_vertex_slice.n_atoms
# 1 for post index
# 1 for n_synapse_types
# 1 for timestep per delay
# 1 for padding
# 4 for Population RNG seed
# 4 for core RNG seed
SYNAPSES_BASE_GENERATOR_SDRAM_USAGE_IN_BYTES = (
    1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 1 + 4 + 4) * BYTES_PER_WORD

DIRECT_MATRIX_HEADER_COST_BYTES = 1 * BYTES_PER_WORD

# Value to use when there is no region
INVALID_REGION_ID = 0xFFFFFFFF


class SynapseRegions(NamedTuple):
    """
    Indices of regions of synapse-implementing binaries.
    """
    synapse_params: int
    pop_table: int
    synaptic_matrix: int
    synapse_dynamics: int
    structural_dynamics: int
    bitfield_filter: int
    connection_builder: int


class SynapseRegionReferences(NamedTuple):
    """
    Indices of regions of synapse-implementing binaries.
    """
    synapse_params: Optional[int] = None
    pop_table: Optional[int] = None
    synaptic_matrix: Optional[int] = None
    synapse_dynamics: Optional[int] = None
    structural_dynamics: Optional[int] = None
    bitfield_filter: Optional[int] = None
    connection_builder: Optional[int] = None


@dataclass(frozen=True)
class AppKeyInfo(object):
    """
    An object which holds an application key and mask along with the other
    details.
    """

    #: The application-level key
    app_key: int
    #: The application-level mask
    app_mask: int
    #: The mask to get the core from the key
    core_mask: int
    #: The shift to get the core from the key
    core_shift: int
    #: The neurons in each core (except possibly the last)
    n_neurons: int
    #: The number of colour bits sent
    n_colour_bits: int

    @property
    def key_and_mask(self) -> BaseKeyAndMask:
        """
        The key and mask as an object.
        """
        return BaseKeyAndMask(self.app_key, self.app_mask)


class SynapticMatrices(object):
    """
    Handler of synaptic matrices for a core of a population vertex.
    """

    __slots__ = (
        # The number of synapse types received
        "__n_synapse_types",
        # The region identifiers
        "__regions",
        # The sub-matrices for each incoming edge
        "__matrices",
        # The address within the synaptic matrix region after the last matrix
        # was written
        "__host_generated_block_addr",
        # The address within the synaptic matrix region after the last
        # generated matrix will be written
        "__on_chip_generated_block_addr",
        # Determine if any of the matrices can be generated on the machine
        "__gen_on_machine",
        # Number of bits to use for neuron IDs
        "__max_atoms_per_core",
        # The stored master population table data
        "__master_pop_data",
        # The stored generated data
        "__generated_data",
        # The size needed for generated data
        "__generated_data_size",
        # The matrices that need to be generated on host
        "__on_host_matrices",
        # The matrices that have been generated on machine
        "__on_machine_matrices",
        # The application vertex
        "__app_vertex",
        # The weight scales
        "__weight_scales",
        # The size of all synaptic blocks added together
        "__all_syn_block_sz",
        # Whether data generation has already happened
        "__data_generated",
        # The size of the bit field data to be allocated
        "__bit_field_size",
        # The bit field key map generated
        "__bit_field_key_map",
        # The maximum generated data, for calculating timeouts
        "__max_gen_data")

    def __init__(
            self, app_vertex: PopulationVertex,
            regions: SynapseRegions, max_atoms_per_core: int,
            weight_scales: WeightScales, all_syn_block_sz: int):
        """
        :param app_vertex:
        :param regions: The synapse regions to use
        :param max_atoms_per_core:
        :param all_syn_block_sz:
        """
        self.__app_vertex = app_vertex
        self.__regions = regions
        self.__n_synapse_types = app_vertex.neuron_impl.get_n_synapse_types()
        self.__max_atoms_per_core = max_atoms_per_core
        self.__weight_scales = weight_scales
        self.__all_syn_block_sz = all_syn_block_sz

        # Map of (app_edge, synapse_info) to SynapticMatrixApp
        self.__matrices: Dict[
            Tuple[ProjectionApplicationEdge, SynapseInformation],
            SynapticMatrixApp] = dict()

        # Store locations of synaptic data and generated data
        self.__host_generated_block_addr = 0
        self.__on_chip_generated_block_addr = 0

        # Determine whether to generate on machine
        self.__gen_on_machine = False
        self.__data_generated = False
        self.__max_gen_data = 0
        self.__on_host_matrices: List[SynapticMatrixApp] = []
        self.__on_machine_matrices: List[SynapticMatrixApp] = []
        self.__generated_data: Optional[NDArray[uint32]] = None
        self.__generated_data_size = 0
        self.__master_pop_data: Optional[NDArray[uint32]] = None
        self.__bit_field_size = 0
        self.__bit_field_key_map: Optional[NDArray[uint32]] = None

    @property
    def max_gen_data(self) -> int:
        """
        The maximum amount of data to be generated for the synapses.
        """
        return self.__max_gen_data

    @property
    def bit_field_size(self) -> int:
        """
        The size of the bit field data.
        """
        return self.__bit_field_size

    @property
    def host_generated_block_addr(self) -> int:
        """
        The address within the synaptic region after the last block
        written by the on-host synaptic generation, i.e. the start of
        the space that can be overwritten provided the synapse expander
        is run again.
        """
        return self.__host_generated_block_addr

    @property
    def on_chip_generated_matrix_size(self) -> int:
        """
        The size of the space used by the generated matrix, i.e. the
        space that can be overwritten provided the synapse expander
        is run again.
        """
        return (self.__on_chip_generated_block_addr -
                self.__host_generated_block_addr)

    def generate_data(self) -> None:
        """
        Generates the data if it has not already been done.
        """
        # If the data has already been generated, stop
        if self.__data_generated:
            return
        self.__data_generated = True

        # If there are no synapses, there is nothing to do!
        if self.__all_syn_block_sz == 0:
            return

        # Track writes inside the synaptic matrix region:
        block_addr = 0

        # Set up the master population table
        poptable = MasterPopTableAsBinarySearch()
        poptable.initialise_table()

        # Set up other lists
        self.__on_host_matrices = list()
        self.__on_machine_matrices = list()
        generated_data: List[NDArray[uint32]] = list()

        # Keep on-machine generated blocks together at the end
        self.__generated_data_size = (
            SYNAPSES_BASE_GENERATOR_SDRAM_USAGE_IN_BYTES +
            (self.__n_synapse_types * DataType.U3232.size))

        # For each incoming machine vertex, reserve pop table space
        for proj in self.__app_vertex.incoming_projections:
            # pylint: disable=protected-access
            app_edge = proj._projection_edge
            synapse_info = proj._synapse_information
            if is_sdram_poisson_source(app_edge):
                continue
            app_key_info = self.__app_key_and_mask(app_edge, synapse_info)
            d_app_key_info = self.__delay_app_key_and_mask(
                app_edge, synapse_info)
            app_matrix = SynapticMatrixApp(
                synapse_info, app_edge, self.__n_synapse_types,
                self.__regions.synaptic_matrix, self.__max_atoms_per_core,
                self.__all_syn_block_sz, app_key_info, d_app_key_info,
                self.__weight_scales)
            self.__matrices[app_edge, synapse_info] = app_matrix

            # If we can generate on machine, store until end
            if synapse_info.may_generate_on_machine():
                self.__on_machine_matrices.append(app_matrix)
            else:
                block_addr = app_matrix.reserve_matrices(block_addr, poptable)
                self.__on_host_matrices.append(app_matrix)

        self.__host_generated_block_addr = block_addr

        # Now add the blocks on machine to keep these all together
        self.__max_gen_data = 0
        for app_matrix in self.__on_machine_matrices:
            block_addr = app_matrix.reserve_matrices(block_addr, poptable)
            gen_data = app_matrix.get_generator_data()
            self.__generated_data_size += gen_data.size
            generated_data.extend(gen_data.gen_data)
            self.__max_gen_data += app_matrix.gen_size
        if generated_data:
            self.__gen_on_machine = True
            self.__generated_data = numpy.concatenate(generated_data)
        else:
            self.__gen_on_machine = True
            self.__generated_data = numpy.zeros(0, dtype=uint32)

        self.__on_chip_generated_block_addr = block_addr

        # Store the master pop table
        self.__master_pop_data = poptable.get_pop_table_data()

        # Store bit field data
        self.__bit_field_size = get_sdram_for_bit_field_region(
            self.__app_vertex.incoming_projections)
        self.__bit_field_key_map = get_bitfield_key_map_data(
            self.__app_vertex.incoming_projections)
        self.__generated_data_size += (
            len(self.__bit_field_key_map) * BYTES_PER_WORD)

    def __write_pop_table(self, spec: DataSpecificationBase,
                          poptable_ref: Optional[int] = None) -> None:
        assert self.__master_pop_data is not None
        master_pop_table_sz = len(self.__master_pop_data) * BYTES_PER_WORD
        spec.reserve_memory_region(
            region=self.__regions.pop_table, size=master_pop_table_sz,
            label='PopTable', reference=poptable_ref)
        spec.switch_write_focus(region=self.__regions.pop_table)
        spec.write_array(self.__master_pop_data)

    def write_synaptic_data(
            self, spec: DataSpecificationBase, post_vertex_slice: Slice,
            references: SynapseRegionReferences) -> None:
        """
        Write the synaptic data for all incoming projections.

        :param spec:
            The spec to write to
        :param post_vertex_slice:
            The slice of the post-vertex the matrix is for
        :param references:
            Regions which are referenced; each region which is not referenced
            can be `None`.
        """
        spec.comment(
            "\nWriting Synaptic Matrix and Master Population Table:\n")

        # Write the pop table
        self.__write_pop_table(spec, references.pop_table)

        # Get the on-host data to be written
        block_addr = 0
        data_to_write: List[NDArray[uint32]] = list()
        for matrix in self.__on_host_matrices:
            block_addr = matrix.append_matrix(
                post_vertex_slice, data_to_write, block_addr)

        # Write on-host data
        spec.reserve_memory_region(
            region=self.__regions.synaptic_matrix,
            size=self.__all_syn_block_sz, label='SynBlocks',
            reference=references.synaptic_matrix)
        if data_to_write:
            spec.switch_write_focus(self.__regions.synaptic_matrix)
            spec.write_array(numpy.concatenate(data_to_write))

        self.__write_synapse_expander_data_spec(
            spec, post_vertex_slice, references.connection_builder)

        write_bitfield_init_data(
            spec, self.__regions.bitfield_filter, self.__bit_field_size,
            references.bitfield_filter)

    def __write_synapse_expander_data_spec(
            self, spec: DataSpecificationBase, post_vertex_slice: Slice,
            connection_builder_ref: Optional[int] = None) -> None:
        """
        Write the data spec for the synapse expander.

        :param spec: The specification to write to
        """
        if self.__generated_data is None:
            if connection_builder_ref is not None:
                # If there is a reference, we still need a region to create
                spec.reserve_memory_region(
                    region=self.__regions.connection_builder,
                    size=4, label="ConnectorBuilderRegion",
                    reference=connection_builder_ref)
            return

        assert self.__bit_field_key_map is not None

        spec.reserve_memory_region(
            region=self.__regions.connection_builder,
            size=self.__generated_data_size, label="ConnectorBuilderRegion",
            reference=connection_builder_ref)
        spec.switch_write_focus(self.__regions.connection_builder)
        spec.write_value(self.__regions.synaptic_matrix)
        spec.write_value(self.__regions.pop_table)
        spec.write_value(self.__regions.bitfield_filter)
        if isinstance(self.__app_vertex.synapse_dynamics,
                      AbstractSynapseDynamicsStructural):
            spec.write_value(self.__regions.structural_dynamics)
        else:
            spec.write_value(INVALID_REGION_ID)
        spec.write_value(len(self.__on_machine_matrices))
        spec.write_value(post_vertex_slice.lo_atom)
        spec.write_value(post_vertex_slice.n_atoms)
        spec.write_value(0)  # TODO: The index if needed
        spec.write_value(self.__n_synapse_types)
        spec.write_value(DataType.S1615.encode_as_int(
            SpynnakerDataView.get_simulation_time_step_per_ms()))
        # Per-Population RNG
        spec.write_array(self.__app_vertex.pop_seed)
        # Per-Core RNG
        spec.write_array(self.__app_vertex.core_seed(post_vertex_slice))
        for w in self.__weight_scales:
            # if the weights are high enough and the population size large
            # enough, then weight_scales < 1 will result in a zero scale
            # if converted to an int, so we use U3232 here instead (as there
            # can be scales larger than U1616.max in conductance-based models)
            dtype = DataType.U3232
            spec.write_value(data=float(min(dtype.max, w)), data_type=dtype)

        spec.write_array(self.__generated_data)
        spec.write_array(self.__bit_field_key_map)

    def __get_app_key_and_mask(
            self, r_info: AppVertexRoutingInfo, n_stages: int,
            pre_vertex: PopulationApplicationVertex,
            partition_id: str) -> AppKeyInfo:
        """
        Get a key and mask for an incoming application vertex as a whole.

        :param r_info: The routing information for the vertex
        :param n_stages: The number of delay stages
        :param pre_vertex: The pre-vertex
        :param partition_id: The partition identifier
        """
        if isinstance(pre_vertex, ApplicationVirtualVertex):
            mask_size = 0
            core_mask = 0
            n_atoms = 0
        else:
            # Find the part that is just for the core
            mask_size = r_info.n_bits_atoms
            core_mask = (2 ** allocator_bits_needed(
                len(r_info.vertex.splitter.get_out_going_vertices(
                    partition_id)))) - 1
            n_atoms = min(pre_vertex.get_max_atoms_per_core(),
                          pre_vertex.n_atoms)

        return AppKeyInfo(
            app_key=r_info.key, app_mask=r_info.mask, core_mask=core_mask,
            core_shift=mask_size, n_neurons=n_atoms * n_stages,
            n_colour_bits=pre_vertex.n_colour_bits)

    def __app_key_and_mask(
            self, app_edge: ProjectionApplicationEdge,
            s_info: SynapseInformation) -> Optional[AppKeyInfo]:
        """
        Get a key and mask for an incoming application vertex as a whole.

        :param app_edge:
            The application edge to get the key and mask of
        :param s_info:
            The synapse information of the projection
        """
        routing_info = SpynnakerDataView.get_routing_infos()
        r_info = routing_info.get_info_from(
            app_edge.pre_vertex, s_info.partition_id)
        assert isinstance(r_info, AppVertexRoutingInfo)
        return self.__get_app_key_and_mask(
            r_info, 1, app_edge.pre_vertex, s_info.partition_id)

    def __delay_app_key_and_mask(
            self, app_edge: ProjectionApplicationEdge,
            s_info: SynapseInformation) -> Optional[AppKeyInfo]:
        """
        Get a key and mask for a whole incoming delayed application
        vertex, or return `None` if no delay edge exists.

        :param app_edge:
            The application edge to get the key and mask of
        :param s_info:
            The synapse information of the projection
        """
        delay_edge = app_edge.delay_edge
        if delay_edge is None:
            return None
        routing_info = SpynnakerDataView.get_routing_infos()
        r_info = routing_info.get_info_from(
            delay_edge.pre_vertex, s_info.partition_id)
        assert isinstance(r_info, AppVertexRoutingInfo)

        # We use the app_edge pre-vertex max atoms here as the delay vertex
        # is split according to this
        return self.__get_app_key_and_mask(
            r_info, app_edge.n_delay_stages, app_edge.pre_vertex,
            s_info.partition_id)

    def get_connections_from_machine(
            self, placement: Placement, app_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> Sequence[NDArray]:
        """
        Get the synaptic connections from the machine.

        :param placement:
            Where the vertices are on the machine
        :param app_edge:
            The application edge of the projection
        :param synapse_info:
            The synapse information of the projection
        :return: A list of arrays of connections, each with dtype
            :py:const:`~.NUMPY_CONNECTORS_DTYPE`
        """
        if is_sdram_poisson_source(app_edge):
            return cast(SpikeSourcePoissonVertex, app_edge.pre_vertex)\
                .read_connections(synapse_info)
        matrix = self.__matrices[app_edge, synapse_info]
        return matrix.get_connections(placement)

    def read_generated_connection_holders(self, placement: Placement) -> None:
        """
        Fill in any pre-run connection holders for data which is generated
        on the machine, after it has been generated.

        :param placement:
            where the data is to be read from
        """
        for matrix in self.__on_machine_matrices:
            matrix.read_generated_connection_holders(placement)

    @property
    def gen_on_machine(self) -> bool:
        """
        Whether any matrices need to be generated on the machine.
        """
        return self.__gen_on_machine

    def get_index(self, app_edge: ProjectionApplicationEdge,
                  synapse_info: SynapseInformation) -> int:
        """
        :param app_edge:
            The application edge of the projection
        :param synapse_info:
            The synapse information of the projection
        :returns: The index of an incoming projection in the population table.
        """
        matrix = self.__matrices[app_edge, synapse_info]
        return matrix.get_index()

    def get_download_regions(
            self, placement: Placement) -> List[Tuple[int, int, int]]:
        """
        Get the regions that need to be downloaded.

        :param placement: The placement of the vertex

        :return: The index, the start address and the size of the regions
        """
        start_index = 0
        regions = list()
        for matrix in self.__matrices.values():
            mat_regions = matrix.get_download_regions(placement, start_index)
            regions.extend(mat_regions)
            start_index += len(mat_regions)
        return regions



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable

import numpy
from numpy import uint16, floating
from numpy.typing import ArrayLike, NDArray

from spinn_utilities.overrides import overrides

from spinn_front_end_common.interface.ds import (
    DataType, DataSpecificationBase)
from spinn_front_end_common.utilities.constants import (
    BYTES_PER_WORD, BYTES_PER_SHORT)

from .abstract_formation import AbstractFormation

# 6 32-bit words (grid_x, grid_y, grid_x_recip, grid_y_recep, ff_prob_size,
#                 lat_prob_size)
_PARAMS_SIZE_IN_BYTES = 6 * BYTES_PER_WORD


class DistanceDependentFormation(AbstractFormation):
    """
    Formation rule that depends on the physical distance between neurons.
    """

    __slots__ = (
        "__grid",
        "__p_form_forward",
        "__sigma_form_forward",
        "__p_form_lateral",
        "__sigma_form_lateral",
        "__ff_distance_probabilities",
        "__lat_distance_probabilities")

    def __init__(
            self, grid: ArrayLike = (16, 16), p_form_forward: float = 0.16,
            sigma_form_forward: float = 2.5, p_form_lateral: float = 1.0,
            sigma_form_lateral: float = 1.0):
        """
        :param grid: (x, y) dimensions of the grid of distance
        :param p_form_forward:
            The peak probability of formation on feed-forward connections
        :param sigma_form_forward:
            The spread of probability with distance of formation on
            feed-forward connections
        :param p_form_lateral:
            The peak probability of formation on lateral connections
        :param sigma_form_lateral:
            The spread of probability with distance of formation on
            lateral connections
        """
        self.__grid = numpy.asarray(grid, dtype=int)
        self.__p_form_forward = p_form_forward
        self.__sigma_form_forward = sigma_form_forward
        self.__p_form_lateral = p_form_lateral
        self.__sigma_form_lateral = sigma_form_lateral

        self.__ff_distance_probabilities = \
            self.generate_distance_probability_array(
                self.__p_form_forward, self.__sigma_form_forward)
        self.__lat_distance_probabilities = \
            self.generate_distance_probability_array(
                self.__p_form_lateral, self.__sigma_form_lateral)

    @property
    @overrides(AbstractFormation.vertex_executable_suffix)
    def vertex_executable_suffix(self) -> str:
        return "_distance"

    @overrides(AbstractFormation.get_parameters_sdram_usage_in_bytes)
    def get_parameters_sdram_usage_in_bytes(self) -> int:
        return (_PARAMS_SIZE_IN_BYTES +
                len(self.__ff_distance_probabilities) * BYTES_PER_SHORT +
                len(self.__lat_distance_probabilities) * BYTES_PER_SHORT)

    def generate_distance_probability_array(
            self, probability: float, sigma: float) -> NDArray[uint16]:
        """
        Generate the exponentially decaying probability LUTs.

        :param probability: peak probability
        :param sigma: spread
        :return: distance-dependent probabilities
        """
        euclidian_distances = numpy.ones(self.__grid ** 2) * numpy.nan
        for row in range(euclidian_distances.shape[0]):
            for column in range(euclidian_distances.shape[1]):
                if self.__grid[0] > 1:
                    pre = (row // self.__grid[0], row % self.__grid[1])
                    post = (column // self.__grid[0], column % self.__grid[1])
                else:
                    pre = (0, row % self.__grid[1])
                    post = (0, column % self.__grid[1])

                # TODO Make distance metric "type" controllable
                euclidian_distances[row, column] = self.distance(
                    pre, post, metric='euclidian')
        largest_squared_distance = numpy.max(euclidian_distances ** 2)
        squared_distances = numpy.arange(largest_squared_distance + 1)
        raw_probabilities = probability * (
            numpy.exp(-squared_distances / (2 * sigma ** 2)))
        quantised_probabilities = raw_probabilities * ((2 ** 16) - 1)
        # Quantize probabilities and cast as uint16 / short
        unfiltered_probabilities = quantised_probabilities.astype(uint16)
        # Only return probabilities which are non-zero
        filtered_probabilities = unfiltered_probabilities[
            unfiltered_probabilities > 0]
        if filtered_probabilities.size % BYTES_PER_SHORT != 0:
            filtered_probabilities = numpy.concatenate(
                (filtered_probabilities,
                 numpy.zeros(filtered_probabilities.size % BYTES_PER_SHORT,
                             dtype=uint16)))

        return filtered_probabilities

    def distance(self, x0: ArrayLike, x1: ArrayLike,
                 metric: str) -> NDArray[floating]:
        """
        Compute the distance between points x0 and x1 place on the grid
        using periodic boundary conditions.

        :param x0: first point in space
        :param x1: second point in space
        :param metric:
            distance metric, i.e. ``euclidian`` or ``manhattan`` or
            ``equidistant``
        :return: the distance
        """
        delta = numpy.abs(numpy.asarray(x0) - numpy.asarray(x1))
        if (delta[0] > self.__grid[0] * .5) and self.__grid[0] > 0:
            delta[0] -= self.__grid[0]

        if (delta[1] > self.__grid[1] * .5) and self.__grid[1] > 0:
            delta[1] -= self.__grid[1]

        if metric == 'manhattan':
            return numpy.abs(delta).sum(axis=-1)
        elif metric == 'equidistant':
            p = 4
            exponents = numpy.power(delta, [p] * delta.size)
            return numpy.floor(numpy.power(exponents.sum(axis=-1), [1. / p]))
        return numpy.sqrt((delta ** 2).sum(axis=-1))

    @overrides(AbstractFormation.write_parameters)
    def write_parameters(self, spec: DataSpecificationBase) -> None:
        spec.write_array(self.__grid)
        # Work out the reciprocal, but zero them if >= 1 as they are not
        # representable as S031 in that case, and not used anyway
        recip = 1 / self.__grid
        recip[recip >= 1.0] = 0
        spec.write_array(DataType.S031.encode_as_numpy_int_array(recip))
        spec.write_value(len(self.__ff_distance_probabilities))
        spec.write_value(len(self.__lat_distance_probabilities))
        spec.write_array(self.__ff_distance_probabilities.view("<u4"))
        spec.write_array(self.__lat_distance_probabilities.view("<u4"))

    @overrides(AbstractFormation.get_parameter_names)
    def get_parameter_names(self) -> Iterable[str]:
        return ("grid", "p_form_forward", "sigma_form_forward",
                "p_form_lateral", "sigma_form_lateral")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from typing import Optional

from spinn_utilities.overrides import overrides
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spynnaker.pyNN.external_devices_models import (
    AbstractMulticastControllableDevice, SendType)
from spynnaker.pyNN.external_devices_models.push_bot import (
    AbstractPushBotOutputDevice)
from spynnaker.pyNN.protocols import MunichIoSpiNNakerLinkProtocol

# The default timestep to use for first send.  Avoids clashes with other
# control commands.
_DEFAULT_FIRST_SEND_TIMESTEP = 100


class PushBotEthernetDevice(
        AbstractMulticastControllableDevice, metaclass=AbstractBase):
    """
    An arbitrary PushBot device.
    """

    def __init__(
            self, protocol: MunichIoSpiNNakerLinkProtocol,
            device: AbstractPushBotOutputDevice,
            uses_payload: bool, time_between_send: Optional[int],
            first_send_timestep: Optional[int] = _DEFAULT_FIRST_SEND_TIMESTEP):
        """
        :param protocol:
            The protocol instance to get commands from
        :param device:
            The Enum instance of the device to control
        :param uses_payload:
            True if the device uses a payload for control
        :param time_between_send: The timesteps between sending
        :param first_send_timestep: The first timestep to send
        """
        self.__protocol = protocol
        self.__device = device
        self.__uses_payload = uses_payload
        self.__time_between_send = time_between_send
        if time_between_send is None:
            self.__time_between_send = device.time_between_send
        self.__first_send_timestep = first_send_timestep

    @property
    @overrides(AbstractMulticastControllableDevice.device_control_key)
    def device_control_key(self) -> int:
        return self.__device.protocol_property.fget(self.__protocol)

    @property
    @overrides(AbstractMulticastControllableDevice.device_control_partition_id)
    def device_control_partition_id(self) -> str:
        return f"{self.__device.name}_PARTITION_ID"

    @property
    @overrides(AbstractMulticastControllableDevice.device_control_uses_payload)
    def device_control_uses_payload(self) -> bool:
        return self.__uses_payload

    @property
    @overrides(AbstractMulticastControllableDevice.device_control_min_value)
    def device_control_min_value(self) -> float:
        return self.__device.min_value

    @property
    @overrides(AbstractMulticastControllableDevice.device_control_max_value)
    def device_control_max_value(self) -> Decimal:
        return self.__device.max_value

    @property
    @overrides(AbstractMulticastControllableDevice
               .device_control_timesteps_between_sending)
    def device_control_timesteps_between_sending(self) -> Optional[int]:
        return self.__time_between_send

    @property
    @overrides(AbstractMulticastControllableDevice
               .device_control_send_type)
    def device_control_send_type(self) -> SendType:
        return self.__device.send_type

    @property
    @overrides(AbstractMulticastControllableDevice
               .device_control_first_send_timestep)
    def device_control_first_send_timestep(self) -> Optional[int]:
        return self.__first_send_timestep

    @property
    def protocol(self) -> MunichIoSpiNNakerLinkProtocol:
        """
        The protocol instance, for use in the subclass.
        """
        return self.__protocol

    @abstractmethod
    def set_command_protocol(
            self, command_protocol: MunichIoSpiNNakerLinkProtocol) -> None:
        """
        Set the protocol use to send setup and shutdown commands,
        separately from the protocol used to control the device.

        :param command_protocol: The protocol to use for this device
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Sequence, Union, Tuple
from spinn_utilities.overrides import overrides
from spinn_utilities.classproperty import classproperty
from pacman.model.partitioner_splitters import AbstractSplitterCommon
from spynnaker.pyNN.models.abstract_pynn_model import AbstractPyNNModel
from .spike_source_poisson_vertex import SpikeSourcePoissonVertex

_population_parameters = {
    "seed": None, "max_rate": None, "splitter": None,
    "n_colour_bits": None, "neurons_per_core": None}

# Technically, this is ~2900 in terms of DTCM, but is timescale dependent
# in terms of CPU (2900 at 10 times slow down is fine, but not at
# real-time)
DEFAULT_MAX_ATOMS_PER_CORE = 500


class SpikeSourcePoisson(AbstractPyNNModel):
    """
    A model of a Poisson-distributed source of spikes.
    """
    __slots__ = ("__duration", "__rate", "__start")

    default_population_parameters = _population_parameters

    def __init__(self, rate: Union[float, Sequence[float]] = 0.0,
                 start: Union[int, Sequence[int]] = 0,
                 duration: Union[int, Sequence[int], None] = None):
        """
        :param rate:
        :param start:
        :param duration:
        """
        self.__start = start
        self.__duration = duration
        self.__rate = rate

    @classproperty
    def absolute_max_atoms_per_core(  # pylint: disable=no-self-argument
            cls) -> int:
        return DEFAULT_MAX_ATOMS_PER_CORE

    @overrides(AbstractPyNNModel.create_vertex)
    def create_vertex(
            self, n_neurons: int, label: str, *,
            seed: Optional[int] = None, max_rate: Optional[float] = None,
            splitter: Optional[AbstractSplitterCommon] = None,
            neurons_per_core: Optional[Union[int, Tuple[int, ...]]] = None,
            n_colour_bits: Optional[int] = None) -> SpikeSourcePoissonVertex:
        """
        :param label: The name for the vertex.
        :param seed:
        :param max_rate:
        :param splitter: The splitter object needed for this vertex.
            Leave as `None` to delegate the choice of splitter to the selector.
        """
        if neurons_per_core is None:
            neurons_per_core = \
                self.get_model_max_atoms_per_dimension_per_core()
        return SpikeSourcePoissonVertex(
            n_neurons, label, seed, neurons_per_core, self,
            rate=self.__rate, start=self.__start, duration=self.__duration,
            max_rate=max_rate, splitter=splitter, n_colour_bits=n_colour_bits)



# Copyright (c) 2025 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import abstractmethod
from pacman.model.graphs.application import ApplicationVertex


class ColouredApplicationVertex(ApplicationVertex):
    """
    Indicates an application vertex that implements n_colour_bits.
    """

    __slots__ = ()

    @property
    @abstractmethod
    def n_colour_bits(self) -> int:
        """
        The number of colour bits sent by this vertex.

        Assumed 0 unless overridden
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from enum import Enum
from typing import Mapping, Optional, Sequence, Union, TYPE_CHECKING
from typing_extensions import TypeAlias
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_front_end_common.interface.ds import DataType
if TYPE_CHECKING:
    from spynnaker.pyNN.models.populations import Population, PopulationBase
    from spynnaker.pyNN.models.neuron.population_vertex import (
        PopulationVertex)

#: General type of parameters to current sources.
#: Individual parameters will only be one of these!
CurrentParameter: TypeAlias = Union[int, float, Sequence[int], Sequence[float]]


class CurrentSourceIDs(Enum):
    """
    Hashes of the current sources currently supported
    """
    NO_SOURCE = 0
    DC_SOURCE = 1
    AC_SOURCE = 2
    STEP_CURRENT_SOURCE = 3
    NOISY_CURRENT_SOURCE = 4
    N_SOURCES = 4


class AbstractCurrentSource(object, metaclass=AbstractBase):
    """
    A simplified version of the PyNN class, since in most cases we work
    out the actual offset value on the SpiNNaker machine itself based on
    the parameters during the run.
    """
    __slots__ = (
        "__app_vertex",
        "__population")

    def __init__(self) -> None:
        self.__app_vertex: Optional[PopulationVertex] = None
        self.__population: Optional[Population] = None

    def inject_into(self, cells: PopulationBase) -> None:
        """
        Inject this source into the specified population cells.

        :param cells: The cells to inject the source into
        """
        # Call the population method to pass the source in
        cells.inject(self)

    def set_app_vertex(self, vertex: PopulationVertex) -> None:
        """
        Set the application vertex associated with the current source.

        :param vertex: The population vertex
        """
        self.__app_vertex = vertex

    @property
    def app_vertex(self) -> Optional[PopulationVertex]:
        """
        The application vertex associated with the current source.
        """
        return self.__app_vertex

    def set_population(self, population: Population) -> None:
        """
        Set the population associated with the current source.

        :param population:
        """
        self.__population = population

    @property
    def population(self) -> Optional[Population]:
        """
        The population associated with the current source.
        """
        return self.__population

    @abstractmethod
    def set_parameters(self, **parameters: CurrentParameter) -> None:
        """
        Set the current source parameters.

        :param parameters: the parameters to set
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def parameters(self) -> Mapping[str, CurrentParameter]:
        """
        The parameters of the current source.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def parameter_types(self) -> Mapping[str, DataType]:
        """
        The parameter types for the current source.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def current_source_id(self) -> int:
        """
        The ID of the current source.
        """
        raise NotImplementedError

    @abstractmethod
    def get_sdram_usage_in_bytes(self) -> int:
        """
        :returns: The SDRAM usage in bytes of the current source.
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The :py:mod:`spynnaker.pyNN` package contains the front end specifications
and implementation for the PyNN High-level API
(https://neuralensemble.org/trac/PyNN).

This package contains the profile of that code for PyNN 0.9.
"""
# pylint: disable=invalid-name

# common imports
import filecmp
import logging
import os
from typing import (
    Any, Callable, Dict, Iterable, List, Optional, Sequence, Tuple, Type,
    TypedDict, Union, cast)

import numpy as __numpy
from typing_extensions import Literal
from numpy.typing import NDArray

from pyNN import common as pynn_common
from pyNN.common import control as _pynn_control
from pyNN.recording import get_io
from pyNN.random import NumpyRNG
from pyNN.space import (
    Space, Line, Grid2D, Grid3D, Cuboid, Sphere, RandomStructure)
from pyNN.space import distance as _pynn_distance
from neo import Block

from spinn_utilities.exceptions import SimulatorNotSetupException
from spinn_utilities.log import FormatAdapter
from spinn_utilities.logger_utils import warn_once
from spinn_utilities.helpful_functions import is_singleton
from spinn_utilities.socket_address import SocketAddress

from spinn_machine.machine import Machine

from spinn_front_end_common.utilities.exceptions import (
    ConfigurationException)

# Self import to check files if copied into pyNN.spiNNaker
import spynnaker.pyNN as _sim  # pylint: disable=import-self

from spynnaker.pyNN.exceptions import SpynnakerException

from spynnaker.pyNN.random_distribution import RandomDistribution
from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.abstract_pynn_model import AbstractPyNNModel
from spynnaker.pyNN.utilities.constants import SPIKE_PARTITION_ID

# connections
# noinspection PyUnresolvedReferences
from spynnaker.pyNN.models.neural_projections.connectors import (
    AbstractConnector, AllToAllConnector, ArrayConnector, CSAConnector,
    DistanceDependentProbabilityConnector, FixedNumberPostConnector,
    FixedNumberPreConnector, FixedProbabilityConnector,
    FromFileConnector, FromListConnector, IndexBasedProbabilityConnector,
    KernelConnector, MultapseConnector as FixedTotalNumberConnector,
    OneToOneConnector, SmallWorldConnector, ConvolutionConnector,
    PoolDenseConnector)
# synapse structures
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    AbstractStaticSynapseDynamics,
    SynapseDynamicsStatic as StaticSynapse)

# plastic stuff
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    SynapseDynamicsSTDP as
    STDPMechanism, SynapseDynamicsStructuralStatic as
    StructuralMechanismStatic, SynapseDynamicsStructuralSTDP as
    StructuralMechanismSTDP)
from spynnaker.pyNN.models.neuron.plasticity.stdp.weight_dependence import (
    WeightDependenceAdditive as
    AdditiveWeightDependence, WeightDependenceMultiplicative as
    MultiplicativeWeightDependence)
from spynnaker.pyNN.models.neuron.plasticity.stdp.timing_dependence import (
    TimingDependenceSpikePair as
    SpikePairRule)
from spynnaker.pyNN.models.neuron.structural_plasticity.synaptogenesis\
    .partner_selection import (
        LastNeuronSelection, RandomSelection)
from spynnaker.pyNN.models.neuron.structural_plasticity.synaptogenesis\
    .formation import (
        DistanceDependentFormation)
from spynnaker.pyNN.models.neuron.structural_plasticity.synaptogenesis\
    .elimination import (
        RandomByWeightElimination)

# local-only synapses
from spynnaker.pyNN.models.neuron.local_only import (
    LocalOnlyConvolution as Convolution,
    LocalOnlyPoolDense as PoolDense)

# neuron stuff
# noinspection PyUnresolvedReferences
from spynnaker.pyNN.models.neuron.builds.if_cond_exp_base import (
    IFCondExpBase as IF_cond_exp)
# noinspection PyUnresolvedReferences
from spynnaker.pyNN.models.neuron.builds.if_curr_exp_base import (
    IFCurrExpBase as IF_curr_exp)
# noinspection PyUnresolvedReferences
from spynnaker.pyNN.models.neuron.builds.if_curr_alpha import (
    IFCurrAlpha as IF_curr_alpha)
# noinspection PyUnresolvedReferences
from spynnaker.pyNN.models.neuron.builds.if_curr_delta import (
    IFCurrDelta as IF_curr_delta)
# noinspection PyUnresolvedReferences
from spynnaker.pyNN.models.neuron.builds.izk_curr_exp_base import (
    IzkCurrExpBase as Izhikevich)
# noinspection PyUnresolvedReferences
from spynnaker.pyNN.models.spike_source.spike_source_array import (
    SpikeSourceArray)
# noinspection PyUnresolvedReferences
from spynnaker.pyNN.models.spike_source.spike_source_poisson import (
    SpikeSourcePoisson)

# pops
# noinspection PyUnresolvedReferences
from spynnaker.pyNN.models.populations import (
    Assembly, Population, PopulationView, IDMixin, PopulationBase)

# projection
# noinspection PyUnresolvedReferences
from spynnaker.pyNN.models.projection import Projection as SpiNNakerProjection

# current sources
# noinspection PyUnresolvedReferences
from spynnaker.pyNN.models.current_sources import (
    DCSource, ACSource, StepCurrentSource, NoisyCurrentSource)

from spynnaker.pyNN import external_devices
from spynnaker.pyNN import extra_models

from spynnaker.pyNN.setup_pynn import setup_pynn

# big stuff
from spynnaker.pyNN.spinnaker import SpiNNaker

from spynnaker._version import __version__  # NOQA
from spynnaker._version import __version_name__  # NOQA
from spynnaker._version import __version_month__  # NOQA
from spynnaker._version import __version_year__  # NOQA


#: The timestep to use of "auto" is specified as a timestep
SPYNNAKER_AUTO_TIMESTEP = 1.0

logger = FormatAdapter(logging.getLogger(__name__))

__all__ = [
    # PyNN imports
    'Cuboid', 'distance', 'Grid2D', 'Grid3D', 'Line', 'NumpyRNG',
    'RandomDistribution', 'RandomStructure', 'Space', 'Sphere',

    # connections
    'AllToAllConnector', 'ArrayConnector', 'CSAConnector',
    'DistanceDependentProbabilityConnector', 'FixedNumberPostConnector',
    'FixedNumberPreConnector', 'FixedProbabilityConnector',
    'FromFileConnector', 'FromListConnector', 'IndexBasedProbabilityConnector',
    'FixedTotalNumberConnector', 'KernelConnector', 'OneToOneConnector',
    'SmallWorldConnector', 'ConvolutionConnector', 'PoolDenseConnector',
    # Local-only
    'Convolution', 'PoolDense',
    # synapse structures
    'StaticSynapse',
    # plastic stuff
    'STDPMechanism', 'AdditiveWeightDependence',
    'MultiplicativeWeightDependence', 'SpikePairRule',
    # Structural plasticity by Petrut Bogdan
    'StructuralMechanismStatic', 'StructuralMechanismSTDP',
    'LastNeuronSelection', 'RandomSelection',
    'DistanceDependentFormation', 'RandomByWeightElimination',
    # neuron stuff
    'IF_cond_exp', 'IF_curr_exp', "IF_curr_alpha", "IF_curr_delta",
    'Izhikevich', 'SpikeSourceArray', 'SpikeSourcePoisson',
    # pops
    'Assembly', 'Population', 'PopulationView',
    # projection
    'SpiNNakerProjection',
    # External devices and extra models
    'external_devices', 'extra_models',
    # CurrentSources
    'DCSource', 'ACSource', 'StepCurrentSource', 'NoisyCurrentSource',
    # Stuff that we define
    'end', 'setup', 'run', 'run_until', 'run_for', 'num_processes', 'rank',
    'reset', 'set_number_of_neurons_per_core',
    'set_number_of_synapse_cores', 'set_allow_delay_extensions',
    'Projection',
    'get_current_time', 'create', 'connect', 'get_time_step', 'get_min_delay',
    'get_max_delay', 'initialize', 'list_standard_models', 'name',
    'record', "get_machine"]


class __PynnOperations(TypedDict, total=False):
    run: Callable[[float, Any], float]
    run_until: Callable[[float, Any], float]
    get_current_time: Callable[[], float]
    get_time_step: Callable[[], float]
    get_max_delay: Callable[[], int]
    get_min_delay: Callable[[], int]
    num_processes: Callable[[], int]
    rank: Callable[[], int]
    reset: Callable[[Dict[str, Any]], None]
    create: Callable[
        [Union[Type, AbstractPyNNModel], Optional[Dict[str, Any]], int],
        Population]
    connect: Callable[
        [Population, Population, float, Optional[float], Optional[str], int,
         Optional[NumpyRNG]], None]
    record: Callable[
        [Union[str, Sequence[str]], PopulationBase, str, Optional[float],
         Optional[Dict[str, Any]]], Block]


# Dynamically-extracted operations from PyNN
__pynn: __PynnOperations = {}
# Cache of the simulator created by setup
__simulator: Optional[SpiNNaker] = None


# Patch the bugs in the PyNN documentation... Ugh!
def distance(src_cell: IDMixin, tgt_cell: IDMixin,
             mask: Optional[NDArray] = None,
             scale_factor: float = 1.0, offset: float = 0.0,
             periodic_boundaries: Optional[Tuple[
                 Optional[Tuple[int, int]]]] = None) -> float:
    """
    :param src_cell: Measure from this cell
    :param tgt_cell: To this cell
    :param mask:
        allows only certain dimensions to be considered, e.g.:

        * to ignore the z-dimension, use ``mask=array([0,1])``
        * to ignore y, ``mask=array([0,2])``
        * to just consider z-distance, ``mask=array([2])``
    :param scale_factor:
        allows for different units in the pre- and post-position
        (the post-synaptic position is multiplied by this quantity).
    :param offset:
    :param periodic_boundaries:
    :returns: The Euclidean distance between two cells.
    """
    return _pynn_distance(
        src_cell, tgt_cell, mask, scale_factor, offset, periodic_boundaries)


def setup(timestep: Optional[Union[float, Literal["auto"]]] = None,
          min_delay: Union[float, Literal["auto"]] = (
              _pynn_control.DEFAULT_MIN_DELAY),
          max_delay: Optional[Union[float, Literal["auto"]]] = None,
          database_socket_addresses: Optional[Iterable[SocketAddress]] = None,
          time_scale_factor: Optional[int] = None,
          n_chips_required: Optional[int] = None,
          n_boards_required: Optional[int] = None,
          **extra_params: Any) -> int:
    """
    The main method needed to be called to make the PyNN 0.8 setup.
    Needs to be called before any other function

    :param timestep:
        the time step of the simulations in microseconds;
        if `None`, the configuration value is used
    :param min_delay: the minimum delay of the simulation
    :param max_delay: Ignored and logs a warning if provided
    :param database_socket_addresses: the sockets used by external devices
        for the database notification protocol
    :param time_scale_factor: multiplicative factor to the machine time step
        (does not affect the neuron models accuracy)
    :param n_chips_required:
        Deprecated! Use n_boards_required instead.
        Must be `None` if n_boards_required specified.
    :param n_boards_required:
        if you need to be allocated a machine (for spalloc) before building
        your graph, then fill this in with a general idea of the number of
        boards you need so that the spalloc system can allocate you a machine
        big enough for your needs.
    :param extra_params: other keyword arguments used to configure PyNN
    :return: MPI rank (always 0 on SpiNNaker)
    :raises \
        ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
        if both ``n_chips_required`` and ``n_boards_required`` are used.
    """
    # pylint: disable=global-statement
    # Check for "auto" values and None
    global __simulator
    if timestep is None:
        logger.warning(
            f"The default PyNN timestep of {_pynn_control.DEFAULT_TIMESTEP} "
            "is less than 1(ms) that SpyNNaker is designed for. "
            "Consider including a timestep in your setup call.")
        timestep = float(_pynn_control.DEFAULT_TIMESTEP)
    elif timestep == "auto":
        timestep = SPYNNAKER_AUTO_TIMESTEP
    if min_delay == "auto":
        min_delay = timestep
    if max_delay:
        logger.warning(
            "max_delay is not supported by sPyNNaker so will be ignored")

    # setup PyNN common stuff
    pynn_common.setup(timestep, min_delay, **extra_params)

    # create stuff simulator
    if SpynnakerDataView.is_setup():
        logger.warning("Calling setup a second time causes the previous "
                       "simulator to be stopped and cleared.")
        # if already exists, kill and rebuild
        try:
            assert __simulator is not None
            __simulator.clear()
        except Exception:  # pylint: disable=broad-except
            logger.exception("Error forcing previous simulation to clear")

    # create the main object for all stuff related software
    __simulator = SpiNNaker(
        time_scale_factor=time_scale_factor, timestep=timestep,
        min_delay=min_delay,
        n_chips_required=n_chips_required,
        n_boards_required=n_boards_required)
    # pylint: disable=protected-access
    external_devices._set_simulator(__simulator)

    # warn about kwargs arguments
    if extra_params:
        logger.warning("Extra params {} have been applied to the setup "
                       "command which we do not consider", extra_params)

    # get overloaded functions from PyNN in relation of our simulator object
    _create_overloaded_functions(__simulator)
    SpynnakerDataView.add_database_socket_addresses(database_socket_addresses)
    return rank()


def name() -> str:
    """
    :returns: The name of the simulator.
    """
    return SpynnakerDataView.get_sim_name()


def Projection(
        presynaptic_population: Population,
        postsynaptic_population: Population,
        connector: AbstractConnector,
        synapse_type: Optional[AbstractStaticSynapseDynamics] = None,
        source: None = None, receptor_type: str = "excitatory",
        space: Optional[Space] = None, label: Optional[str] = None,
        download_synapses: bool = False,
        partition_id: str = SPIKE_PARTITION_ID) -> SpiNNakerProjection:
    """
    Used to support PEP 8 spelling correctly.

    :param presynaptic_population: the source pop
    :param postsynaptic_population: the destination population
    :param connector: the connector type
    :param synapse_type: the synapse type
    :param source: Unsupported; must be ``None``
    :param receptor_type: the receptor type
    :param space: the space object
    :param label: the label
    :param download_synapses: whether to download synapses
    :param partition_id: the partition id to use for the projection
    :return: a projection object for SpiNNaker
    """
    return SpiNNakerProjection(
        pre_synaptic_population=presynaptic_population,
        post_synaptic_population=postsynaptic_population, connector=connector,
        synapse_type=synapse_type, source=source, receptor_type=receptor_type,
        space=space, label=label, download_synapses=download_synapses,
        partition_id=partition_id)


def _create_overloaded_functions(spinnaker_simulator: SpiNNaker) -> None:
    """
    Creates functions that the main PyNN interface supports
    (given from PyNN)

    :param spinnaker_simulator: the simulator object we use underneath
    """
    # overload the failed ones with now valid ones, now that we're in setup
    # phase.
    __pynn["run"], __pynn["run_until"] = pynn_common.build_run(
        spinnaker_simulator)

    __pynn["get_current_time"], __pynn["get_time_step"], \
        __pynn["get_min_delay"], __pynn["get_max_delay"], \
        __pynn["num_processes"], __pynn["rank"] = \
        pynn_common.build_state_queries(spinnaker_simulator)

    __pynn["reset"] = pynn_common.build_reset(spinnaker_simulator)
    __pynn["create"] = pynn_common.build_create(Population)

    __pynn["connect"] = pynn_common.build_connect(
        Projection, FixedProbabilityConnector, StaticSynapse)

    __pynn["record"] = pynn_common.build_record(spinnaker_simulator)


def end(_: Any = True) -> None:
    """
    Cleans up the SpiNNaker machine and software

    :param _: was named `compatible_output`, which we don't care about,
        so is a non-existent parameter
    """
    if SpynnakerDataView.is_shutdown():
        logger.warning("Second call to end ignored")
        return
    try:
        SpynnakerDataView.check_valid_simulator()
    except SimulatorNotSetupException:
        logger.exception("Calling end before setup makes no sense ignoring!")
        return
    assert __simulator is not None, "no current simulator"
    for (population, variables, filename) in \
            __simulator.write_on_end:
        io = get_io(filename)
        population.write_data(io, variables)
    __simulator.write_on_end = []
    __simulator.stop()


def list_standard_models() -> List[str]:
    """
    :returns: A list of all the StandardCellType classes available for this
        simulator.
    """
    return [
        key
        for (key, obj) in globals().items()
        if isinstance(obj, type) and issubclass(obj, AbstractPyNNModel)]


def set_number_of_neurons_per_core(
        neuron_type: Type,
        max_permitted: Optional[Union[int, Tuple[int, ...]]]) -> None:
    """
    Sets a ceiling on the number of neurons of a given model that can be
    placed on a single core.
    This can be overridden by the individual Population.

    The new value can be `None`, meaning that the maximum is the same as
    the number of atoms, an int, meaning all Populations of this model
    must have one dimension, or a tuple of n integers, meaning all
    Populations of this model must have n dimensions.
    If not all Populations of this model have the same number of
    dimensions, it is recommended to set this to `None` here and then
    set the maximum on each Population.

    :param neuron_type: neuron type
    :param max_permitted: the number to set to
    """
    if isinstance(neuron_type, str):
        raise ConfigurationException(
            "set_number_of_neurons_per_core call now expects "
            "neuron_type as a class instead of as a str")
    max_neurons: Optional[Tuple[int, ...]] = None
    if max_permitted is not None:
        if is_singleton(max_permitted):
            max_neurons = (int(max_permitted), )
        else:
            max_perm: Tuple[int, ...] = cast(Tuple[int, ...], max_permitted)
            max_neurons = tuple(int(m) for m in max_perm)

    neuron_type.set_model_max_atoms_per_dimension_per_core(max_neurons)
    if SpynnakerDataView.get_n_populations() > 0:
        warn_once(logger,
                  "set_number_of_neurons_per_core "
                  "only affects Populations not yet made.")


def set_number_of_synapse_cores(
        neuron_type: Type, n_synapse_cores: Optional[int]) -> None:
    """
    Sets the number of synapse cores for a model.

    :param neuron_type: The model implementation
    :param n_synapse_cores:
        The number of synapse cores; 0 to force combined cores, and None to
        allow the system to choose
    """
    neuron_type.set_model_n_synapse_cores(n_synapse_cores)
    if SpynnakerDataView.get_n_populations() > 0:
        warn_once(logger,
                  "set_number_of_synapse_cores "
                  "only affects Populations not yet made.")


def set_allow_delay_extensions(
        neuron_type: Type, allow_delay_extensions: bool) -> None:
    """
    Sets whether to allow delay extensions for a model.

    :param neuron_type: The model implementation
    :param allow_delay_extensions: Whether to allow delay extensions
    """
    neuron_type.set_model_allow_delay_extensions(allow_delay_extensions)
    if SpynnakerDataView.get_n_populations() > 0:
        warn_once(logger,
                  "set_allow_delay_extensions "
                  "only affects Populations not yet made.")


# These methods will defer to PyNN methods if a simulator exists


def connect(pre: Population, post: Population, weight: float = 0.0,
            delay: Optional[float] = None, receptor_type: Optional[str] = None,
            p: int = 1, rng: Optional[NumpyRNG] = None) -> None:
    """
    Builds a projection.

    :param pre: source pop
    :param post: destination pop
    :param weight: weight of the connections
    :param delay: the delay of the connections
    :param receptor_type: excitatory / inhibitory
    :param p: probability
    :param rng: random number generator
    """
    SpynnakerDataView.check_user_can_act()
    __pynn["connect"](pre, post, weight, delay, receptor_type, p, rng)


def create(
        cellclass: Union[Type, AbstractPyNNModel],
        cellparams: Optional[Dict[str, Any]] = None,
        n: int = 1) -> Population:
    """
    Builds a population with certain parameters.

    :param cellclass: population class
    :param cellparams: population parameters.
    :param n: number of neurons
    :returns: A new Population
    """
    SpynnakerDataView.check_user_can_act()
    return __pynn["create"](cellclass, cellparams, n)


def NativeRNG(seed_value: Union[int, List[int], NDArray]) -> None:
    """
    Fixes the random number generator's seed.

    :param seed_value:
    """
    __numpy.random.seed(seed_value)


def get_current_time() -> float:
    """
    Gets the time within the simulation.

    :return: returns the current time
    """
    SpynnakerDataView.check_user_can_act()
    return __pynn["get_current_time"]()


def get_min_delay() -> int:
    """
    The minimum allowed synaptic delay; delays will be clamped to be at
    least this.

    :return: returns the min delay of the simulation
    """
    SpynnakerDataView.check_user_can_act()
    return __pynn["get_min_delay"]()


def get_max_delay() -> int:
    """
    Part of the PyNN API but does not make sense for sPyNNaker as
    different Projection, Vertex splitter combination could have different
    delays they can support.

    Most likely value is timestep * 144

    :raises NotImplementedError: As there is no system wide max_delay
    :returns: In SpyNNaker this method never returns
    """
    raise NotImplementedError(
        "sPyNNaker does not have a system wide max_delay")


def get_time_step() -> float:
    """
    The integration time step.

    :return: get the time step of the simulation (in ms)
    """
    SpynnakerDataView.check_user_can_act()
    return float(__pynn["get_time_step"]())


def initialize(cells: PopulationBase, **initial_values: Any) -> None:
    """
    Sets cells to be initialised to the given values.

    :param cells: the cells to change parameters on
    :param initial_values: the parameters and their values to change
    """
    SpynnakerDataView.check_user_can_act()
    pynn_common.initialize(cells, **initial_values)


def num_processes() -> int:
    """
    The number of MPI processes.

    .. note::
        Always 1 on SpiNNaker, which doesn't use MPI.

    :return: the number of MPI processes
    """
    SpynnakerDataView.check_user_can_act()
    return __pynn["num_processes"]()


def rank() -> int:
    """
    The MPI rank of the current node.

    .. note::
        Always 0 on SpiNNaker, which doesn't use MPI.

    :return: MPI rank
    """
    SpynnakerDataView.check_user_can_act()
    return __pynn["rank"]()


def record(variables: Union[str, Sequence[str]], source: PopulationBase,
           filename: str, sampling_interval: Optional[float] = None,
           annotations: Optional[Dict[str, Any]] = None) -> Block:
    """
    Sets variables to be recorded.

    :param variables: may be either a single variable name or a list of
        variable names. For a given `celltype` class, `celltype.recordable`
        contains a list of variables that can be recorded for that `celltype`.
    :param source: where to record from
    :param filename: file name to write data to
    :param sampling_interval:
        how often to sample the recording, not ignored so far
    :param annotations: the annotations to data writers
    :return: neo object
    """
    SpynnakerDataView.check_user_can_act()
    return __pynn["record"](variables, source, filename, sampling_interval,
                            annotations)


def reset(annotations: Optional[Dict[str, Any]] = None) -> None:
    """
    Resets the simulation to t = 0.

    :param annotations: the annotations to the data objects
    """
    if annotations is None:
        annotations = {}
    SpynnakerDataView.check_user_can_act()
    __pynn["reset"](annotations)


def run(simtime: float, callbacks: Optional[Callable] = None) -> float:
    """
    The run() function advances the simulation for a given number of
    milliseconds.

    :param simtime: time to run for (in milliseconds)
    :param callbacks: callbacks to run
    :return: the actual simulation time that the simulation stopped at
    """
    SpynnakerDataView.check_user_can_act()
    return __pynn["run"](simtime, callbacks)


# left here because needs to be done, and no better place to put it
# (ABS don't like it, but will put up with it)
run_for = run


def run_until(tstop: float) -> float:
    """
    Run until a (simulation) time period has completed.

    :param tstop: the time to stop at (in milliseconds)
    :return: the actual simulation time that the simulation stopped at
    """
    SpynnakerDataView.check_user_can_act()
    return __pynn["run_until"](tstop, None)


def get_machine() -> Machine:
    """
    Get the SpiNNaker machine in use.

    :return: the machine object
    """
    SpynnakerDataView.check_user_can_act()
    assert __simulator is not None
    return __simulator.get_machine()


# Check copy in case being run from pyNN.spiNNaker
indirect = os.path.abspath(_sim.__file__)
direct = __file__
if direct != indirect:
    if not filecmp.cmp(direct, indirect):
        setup_pynn()
        raise SpynnakerException(
            "pyNN.spiNNaker needed updating please restart your script")



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Iterable, List, Optional, Tuple, TYPE_CHECKING

import numpy
from numpy import floating, integer, uint8, uint32
from numpy.typing import NDArray

from pyNN.standardmodels.synapses import StaticSynapse

from spinn_utilities.overrides import overrides

from spinn_front_end_common.interface.ds import DataType, DataSpecificationBase
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.exceptions import (
    SynapticConfigurationException, InvalidParameterType)
from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
    NUMPY_CONNECTORS_DTYPE)
from spynnaker.pyNN.models.neuron.plasticity.stdp.common import (
    STDP_FIXED_POINT_ONE, get_exp_lut_array)
from spynnaker.pyNN.types import WEIGHTS_DELAYS_IN as _Weight

from .abstract_plastic_synapse_dynamics import AbstractPlasticSynapseDynamics
from .abstract_generate_on_machine import (
    AbstractGenerateOnMachine, MatrixGeneratorID)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)
    from spynnaker.pyNN.models.neuron.synapse_io import MaxRowInfo
    from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
        ConnectionsArray)
    from .abstract_synapse_dynamics import AbstractSynapseDynamics

# The targets of neuromodulation
NEUROMODULATION_TARGETS = {
    "reward": 0,
    "punishment": 1
}

# LOOKUP_TAU_C_SIZE = 520
LOOKUP_TAU_C_SHIFT = 4
# LOOKUP_TAU_D_SIZE = 370
LOOKUP_TAU_D_SHIFT = 2


class SynapseDynamicsNeuromodulation(
        AbstractPlasticSynapseDynamics, AbstractGenerateOnMachine):
    """
    Synapses that target a neuromodulation receptor.
    """

    __slots__ = (
        "__tau_c",
        "__tau_d",
        "__tau_c_data",
        "__tau_d_data",
        "__w_min",
        "__w_max")

    def __init__(
            self, weight: _Weight = StaticSynapse.default_parameters['weight'],
            tau_c: float = 1000.0, tau_d: float = 200.0,
            w_min: float = 0.0, w_max: float = 1.0):
        """
        :param weight: The weights or way to generate the weights
        :param tau_c:
        :param tau_d:
        :param w_min:
        :param w_max:
        """
        super().__init__(delay=1, weight=weight)
        self.__tau_c = tau_c
        self.__tau_d = tau_d
        ts = SpynnakerDataView.get_simulation_time_step_ms()
        self.__tau_c_data = get_exp_lut_array(
            ts, self.__tau_c, shift=LOOKUP_TAU_C_SHIFT)
        self.__tau_d_data = get_exp_lut_array(
            ts, self.__tau_d, shift=LOOKUP_TAU_D_SHIFT)
        self.__w_min = w_min
        self.__w_max = w_max

        if w_min < 0 or w_max < 0:
            raise SynapticConfigurationException(
                "Minimum and maximum weights must be >= 0")

    @property
    def tau_c(self) -> float:
        """
        The tau c value passed into the init.
        """
        return self.__tau_c

    @property
    def tau_d(self) -> float:
        """
        The tau d value passed into the init.
        """
        return self.__tau_d

    @property
    def w_min(self) -> float:
        """
        The w min value passed into the init.
        """
        return self.__w_min

    @property
    def w_max(self) -> float:
        """
        The w max value passed into the init.
        """
        return self.__w_max

    @overrides(AbstractPlasticSynapseDynamics.merge)
    def merge(self, synapse_dynamics: AbstractSynapseDynamics
              ) -> AbstractSynapseDynamics:
        # This must replace something that supports neuromodulation,
        # so it can't be the first thing to be merged!
        raise SynapticConfigurationException(
            "Neuromodulation synapses can only be added where an existing"
            " projection has already been added which supports"
            " neuromodulation")

    @overrides(AbstractPlasticSynapseDynamics.is_same_as)
    def is_same_as(self, synapse_dynamics: AbstractSynapseDynamics) -> bool:
        # Shouldn't ever come up, but if it does, it is False!
        return False

    def is_neuromodulation_same_as(
            self, other: SynapseDynamicsNeuromodulation) -> bool:
        """
        Checks that tau c, tau d, w max and w min are all the same.

        :param other:
        :returns: True if all checked values are the same
        """
        return (self.__tau_c == other.tau_c and self.__tau_d == other.tau_d and
                self.__w_min == other.w_min and self.__w_max == other.w_max)

    @overrides(AbstractPlasticSynapseDynamics.get_vertex_executable_suffix)
    def get_vertex_executable_suffix(self) -> str:
        return "izhikevich_neuromodulation_"

    @overrides(AbstractPlasticSynapseDynamics
               .get_parameters_sdram_usage_in_bytes)
    def get_parameters_sdram_usage_in_bytes(
            self, n_neurons: int, n_synapse_types: int) -> int:
        size = BYTES_PER_WORD * 3
        size += BYTES_PER_WORD * len(self.__tau_c_data)
        size += BYTES_PER_WORD * len(self.__tau_d_data)
        return size

    @overrides(AbstractPlasticSynapseDynamics.write_parameters)
    def write_parameters(
            self, spec: DataSpecificationBase, region: int,
            global_weight_scale: float,
            synapse_weight_scales: NDArray[floating]) -> None:
        # Calculate constant component in Izhikevich's model weight update
        # function and write to SDRAM.
        weight_update_component = \
            1 / (-((1.0/self.__tau_c) + (1.0/self.__tau_d)))
        spec.write_value(data=weight_update_component,
                         data_type=DataType.S1615)

        # Write min and max weight
        spec.write_value(data=self.__w_max * global_weight_scale,
                         data_type=DataType.S1615)
        spec.write_value(data=self.__w_min * global_weight_scale,
                         data_type=DataType.S1615)

        # Write the LUT arrays
        spec.write_array(self.__tau_c_data)
        spec.write_array(self.__tau_d_data)

    @overrides(AbstractPlasticSynapseDynamics.get_value)
    def get_value(self, key: str) -> float:
        if hasattr(self, key):
            return getattr(self, key)
        raise InvalidParameterType(
            f"Type {type(self)} does not have parameter {key}")

    @overrides(AbstractPlasticSynapseDynamics.set_value)
    def set_value(self, key: str, value: float) -> None:
        if hasattr(self, key):
            setattr(self, key, value)
        raise InvalidParameterType(
            f"Type {type(self)} does not have parameter {key}")

    @overrides(AbstractPlasticSynapseDynamics
               .get_n_words_for_plastic_connections)
    def get_n_words_for_plastic_connections(self, n_connections: int) -> int:
        # 1 for flags
        pp_size_words = 1
        # 1 or each connection
        fp_size_words = n_connections
        return pp_size_words + fp_size_words

    @overrides(AbstractPlasticSynapseDynamics.get_plastic_synaptic_data)
    def get_plastic_synaptic_data(
            self, connections: ConnectionsArray,
            connection_row_indices: NDArray[integer], n_rows: int,
            n_synapse_types: int,
            max_n_synapses: int, max_atoms_per_core: int) -> Tuple[
                NDArray[uint32], NDArray[uint32], NDArray[uint32],
                NDArray[uint32]]:
        weights = numpy.rint(
            numpy.abs(connections["weight"]) * STDP_FIXED_POINT_ONE)
        fixed_plastic = (
            ((weights.astype(uint32) & 0xFFFF) << 16) |
            (connections["target"] & 0xFFFF))
        fixed_plastic_rows = self.convert_per_connection_data_to_rows(
            connection_row_indices, n_rows,
            fixed_plastic.view(dtype=uint8).reshape((-1, BYTES_PER_WORD)),
            max_n_synapses)

        # It is assumed that all connections have the same synapse type
        is_reward = 0
        synapse_type = 0
        if len(connections) > 0:
            synapse_type = connections[0]["synapse_type"]
            is_reward = synapse_type == NEUROMODULATION_TARGETS["reward"]
        flags = 0x80000000 | (int(is_reward) << 30) | synapse_type

        fp_size = self.get_n_items(fixed_plastic_rows, BYTES_PER_WORD)
        fp_data = numpy.vstack([
            fixed_row.view(uint32) for fixed_row in fixed_plastic_rows])
        pp_data = numpy.full((n_rows, 1), flags, dtype=uint32)
        pp_size = numpy.ones(n_rows, dtype=uint32)

        return fp_data, pp_data, fp_size, pp_size

    @overrides(
        AbstractPlasticSynapseDynamics.get_n_plastic_plastic_words_per_row)
    def get_n_plastic_plastic_words_per_row(
            self, pp_size: NDArray[integer]) -> NDArray[integer]:
        # pp_size is in words, so just return
        return pp_size

    @overrides(
        AbstractPlasticSynapseDynamics.get_n_fixed_plastic_words_per_row)
    def get_n_fixed_plastic_words_per_row(
            self, fp_size: NDArray[integer]) -> NDArray[integer]:
        # fp_size is in words, so just return
        return fp_size

    @overrides(AbstractPlasticSynapseDynamics.get_n_synapses_in_rows)
    def get_n_synapses_in_rows(
            self, pp_size: NDArray[integer],
            fp_size: NDArray[integer]) -> NDArray[integer]:
        # Each fixed-plastic synapse is a word and fp_size is in words so just
        # return it
        return fp_size

    @overrides(AbstractPlasticSynapseDynamics.read_plastic_synaptic_data)
    def read_plastic_synaptic_data(
            self, n_synapse_types: int,
            pp_size: NDArray[uint32], pp_data: List[NDArray[uint32]],
            fp_size: NDArray[uint32], fp_data: List[NDArray[uint32]],
            max_atoms_per_core: int) -> ConnectionsArray:
        data = numpy.concatenate(fp_data)
        connections = numpy.zeros(data.size, dtype=NUMPY_CONNECTORS_DTYPE)
        connections["source"] = numpy.concatenate(
            [numpy.repeat(i, fp_size[i]) for i in range(len(fp_size))])
        connections["target"] = data & 0xFFFF
        connections["weight"] = (data >> 16) & 0xFFFF
        connections["delay"] = 1
        return connections

    @overrides(AbstractPlasticSynapseDynamics.get_parameter_names)
    def get_parameter_names(self) -> Iterable[str]:
        yield 'weight'

    @overrides(AbstractPlasticSynapseDynamics.get_max_synapses)
    def get_max_synapses(self, n_words: int) -> int:
        # One word is static, the rest is for synapses
        return n_words - 1

    @property
    @overrides(AbstractGenerateOnMachine.gen_matrix_id)
    def gen_matrix_id(self) -> int:
        return MatrixGeneratorID.NEUROMODULATION_MATRIX.value

    @overrides(AbstractGenerateOnMachine.gen_matrix_params)
    def gen_matrix_params(
            self, synaptic_matrix_offset: int, delayed_matrix_offset: int,
            app_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation, max_row_info: MaxRowInfo,
            max_pre_atoms_per_core: int,
            max_post_atoms_per_core: int) -> NDArray[uint32]:
        _ = (delayed_matrix_offset, max_pre_atoms_per_core,
             max_post_atoms_per_core)
        synapse_type = synapse_info.synapse_type
        return numpy.array([
            synaptic_matrix_offset, max_row_info.undelayed_max_words,
            max_row_info.undelayed_max_n_synapses, app_edge.pre_vertex.n_atoms,
            synapse_type is NEUROMODULATION_TARGETS["reward"], synapse_type],
            dtype=uint32)

    @property
    @overrides(AbstractGenerateOnMachine.
               gen_matrix_params_size_in_bytes)
    def gen_matrix_params_size_in_bytes(self) -> int:
        return 6 * BYTES_PER_WORD

    @property
    @overrides(AbstractPlasticSynapseDynamics.changes_during_run)
    def changes_during_run(self) -> bool:
        return False

    @property
    @overrides(AbstractPlasticSynapseDynamics.pad_to_length)
    def pad_to_length(self) -> None:
        return None

    @overrides(AbstractPlasticSynapseDynamics.get_synapse_id_by_target)
    def get_synapse_id_by_target(self, target: str) -> Optional[int]:
        return NEUROMODULATION_TARGETS.get(target, None)

    @property
    @overrides(AbstractPlasticSynapseDynamics.is_combined_core_capable)
    def is_combined_core_capable(self) -> bool:
        return False

    @property
    @overrides(AbstractPlasticSynapseDynamics.is_split_core_capable)
    def is_split_core_capable(self) -> bool:
        return True

    @property
    @overrides(AbstractPlasticSynapseDynamics.synapses_per_second)
    def synapses_per_second(self) -> int:
        # This should never end up being requested!
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from enum import Enum
from typing import TYPE_CHECKING

from numpy import uint32
from numpy.typing import NDArray

from spinn_utilities.abstract_base import AbstractBase, abstractmethod

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)
    from spynnaker.pyNN.models.neuron.synapse_io import MaxRowInfo


class MatrixGeneratorID(Enum):
    """
    Type of Matrix in use.
    """
    STATIC_MATRIX = 0
    STDP_MATRIX = 1
    NEUROMODULATION_MATRIX = 2
    CHANGE_WEIGHT_MATRIX = 3


class AbstractGenerateOnMachine(object, metaclass=AbstractBase):
    """
    A synapse dynamics that can be generated on the machine.
    """
    __slots__ = ()

    def generate_on_machine(self) -> bool:
        """
        Determines if this instance should be generated on the machine.

        Default implementation returns True

        :returns: True if the instance should be generated on the machine.
        """
        return True

    @property
    @abstractmethod
    def gen_matrix_id(self) -> int:
        """
        The ID of the on-machine matrix generator.
        """
        raise NotImplementedError

    @abstractmethod
    def gen_matrix_params(
            self, synaptic_matrix_offset: int, delayed_matrix_offset: int,
            app_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation, max_row_info: MaxRowInfo,
            max_pre_atoms_per_core: int, max_post_atoms_per_core: int
            ) -> NDArray[uint32]:
        """
        :returns: Any parameters required by the matrix generator.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def gen_matrix_params_size_in_bytes(self) -> int:
        """
        The size of the parameters of the matrix generator in bytes.
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, List, Literal, Optional, Union
from typing_extensions import TypeAlias

from spinn_utilities.overrides import overrides

from pacman.model.routing_info import BaseKeyAndMask
from pacman.model.graphs.application import ApplicationSpiNNakerLinkVertex

from spinn_front_end_common.abstract_models import (
    AbstractSendMeMulticastCommandsVertex)
from spinn_front_end_common.utility_models import MultiCastCommand

from spynnaker.pyNN.exceptions import SpynnakerException
from spynnaker.pyNN.models.common import PopulationApplicationVertex

# robot with 7 7 1


def get_x_from_robot_retina(key: int) -> int:
    """
    :param key:
    :returns: x coordinate from key
    """
    return (key >> 7) & 0x7f


def get_y_from_robot_retina(key: int) -> int:
    """
    :param key:
    :returns: y coordinate from key
    """
    return key & 0x7f


def get_spike_value_from_robot_retina(key: int) -> int:
    """
    :param key:
    :returns: spike from key
    """
    return (key >> 14) & 0x1


class MunichRetinaDevice(
        ApplicationSpiNNakerLinkVertex, PopulationApplicationVertex,
        AbstractSendMeMulticastCommandsVertex):
    """
    An Omnibot silicon retina device.
    """
    __slots__ = (
        "__fixed_key",
        "__fixed_mask",
        "__is_right")

    # key codes for the robot retina
    _MANAGEMENT_BIT = 0x400
    _MANAGEMENT_MASK = 0xFFFFF800
    _LEFT_RETINA_ENABLE = 0x45
    _RIGHT_RETINA_ENABLE = 0x46
    _LEFT_RETINA_DISABLE = 0x45
    _RIGHT_RETINA_DISABLE = 0x46
    _LEFT_RETINA_KEY_SET = 0x43
    _RIGHT_RETINA_KEY_SET = 0x44

    UP_POLARITY = "UP"
    DOWN_POLARITY = "DOWN"
    MERGED_POLARITY: Literal["MERGED"] = "MERGED"
    _T_POLARITY: TypeAlias = \
        Union[Literal["UP"], Literal["DOWN"], Literal["MERGED"], None]

    #: Select the left retina
    LEFT_RETINA = "LEFT"
    #: Select the right retina
    RIGHT_RETINA = "RIGHT"
    _RETINAS = frozenset((LEFT_RETINA, RIGHT_RETINA))
    _T_RETINAS: TypeAlias = Union[Literal["LEFT"], Literal["RIGHT"]]

    def __init__(
            self, retina_key: int, spinnaker_link_id: int,
            position: _T_RETINAS, label: str = "MunichRetinaDevice",
            polarity: _T_POLARITY = None,
            board_address: Optional[str] = None):
        """
        :param retina_key:
        :param spinnaker_link_id:
            The SpiNNaker link to which the retina is connected
        :param position: ``LEFT`` or ``RIGHT``
        :param label: The optional name of the vertex.
        :param polarity: ``UP``, ``DOWN`` or ``MERGED``
        :param board_address:
        """
        if polarity is None:
            polarity = MunichRetinaDevice.MERGED_POLARITY

        self.__fixed_key = (retina_key & 0xFFFF) << 16
        self.__fixed_mask = 0xFFFF8000
        if polarity == MunichRetinaDevice.UP_POLARITY:
            self.__fixed_key |= 0x4000

        if polarity == MunichRetinaDevice.MERGED_POLARITY:
            # There are 128 x 128 retina "pixels" x 2 polarities
            fixed_n_neurons = 128 * 128 * 2
        else:
            # There are 128 x 128 retina "pixels"
            fixed_n_neurons = 128 * 128
            self.__fixed_mask = 0xFFFFC000

        if position not in self._RETINAS:
            raise SpynnakerException(
                "The external Retina does not recognise this position")
        self.__is_right = position == self.RIGHT_RETINA

        super().__init__(
            n_atoms=fixed_n_neurons, spinnaker_link_id=spinnaker_link_id,
            label=label, board_address=board_address)

    @overrides(ApplicationSpiNNakerLinkVertex.get_fixed_key_and_mask)
    def get_fixed_key_and_mask(self, partition_id: str) -> BaseKeyAndMask:
        return BaseKeyAndMask(self.__fixed_key, self.__fixed_mask)

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.start_resume_commands)
    def start_resume_commands(self) -> Iterable[MultiCastCommand]:
        # change the retina key it transmits with
        # (based off if its right or left)
        key_set_command = self._MANAGEMENT_BIT | (
            self._RIGHT_RETINA_KEY_SET if self.__is_right
            else self._LEFT_RETINA_KEY_SET)

        # to ensure populations receive the correct packets, this needs to be
        # different based on which retina
        key_set_payload = self.__fixed_key if self.__is_right else 0

        yield MultiCastCommand(
            key_set_command, key_set_payload,
            repeat=5, delay_between_repeats=1000)

        # make retina enabled (dependent on if its a left or right retina
        enable_command = self._MANAGEMENT_BIT | (
            self._RIGHT_RETINA_ENABLE if self.__is_right
            else self._LEFT_RETINA_ENABLE)

        yield MultiCastCommand(
            enable_command, payload=1, repeat=5, delay_between_repeats=1000)

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.pause_stop_commands)
    def pause_stop_commands(self) -> Iterable[MultiCastCommand]:
        # disable retina
        disable_command = self._MANAGEMENT_BIT | (
            self._RIGHT_RETINA_DISABLE if self.__is_right
            else self._LEFT_RETINA_DISABLE)

        yield MultiCastCommand(
            disable_command, payload=0, repeat=5, delay_between_repeats=1000)

    @property
    @overrides(AbstractSendMeMulticastCommandsVertex.timed_commands)
    def timed_commands(self) -> List[MultiCastCommand]:
        return []



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import logging
from typing import (
    Any, Collection, Dict, Mapping, Optional, Sequence, TYPE_CHECKING)

import neo  # type: ignore[import]

from spinn_utilities.log import FormatAdapter
from spinn_utilities.logger_utils import warn_once

from spinn_front_end_common.utilities.exceptions import ConfigurationException

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.utilities.neo_buffer_database import NeoBufferDatabase
from spynnaker.pyNN.types import IoDest

if TYPE_CHECKING:
    from spynnaker.pyNN.models.common.types import Names
    from spynnaker.pyNN.models.populations import Population
    from spynnaker.pyNN.models.common import PopulationApplicationVertex

logger = FormatAdapter(logging.getLogger(__name__))


class Recorder(object):
    """
    Object to hold recording behaviour, used by populations.
    """

    __slots__ = (
        "__population",
        "__vertex",
        "__write_to_files_indicators")

    def __init__(
            self, population: Population, vertex: PopulationApplicationVertex):
        """
        :param population:
            the population to record for
        :param vertex:
            the SpiNNaker graph vertex used by the population
        """
        self.__population = population
        self.__vertex = vertex

        # file flags, allows separate files for the recorded variables
        self.__write_to_files_indicators: Dict[str, IoDest] = {
            'spikes': None,
            'gsyn_exc': None,
            'gsyn_inh': None,
            'v': None}

    @property
    def write_to_files_indicators(self) -> Mapping[str, IoDest]:
        """
        What variables should be written to files, and where should they
        be written.
        """
        return self.__write_to_files_indicators

    def record(
            self, variables: Names, to_file: IoDest,
            sampling_interval: Optional[int],
            indexes: Optional[Collection[int]]) -> None:
        """
        Turns on (or off) recording.

        :param variables: either a single variable name or a list of variable
            names. For a given `celltype` class, `celltype.recordable` contains
            a list of variables that can be recorded for that `celltype`.
            Can also be ``None`` to reset the list of variables.
        :param to_file: a file to automatically record to (optional).
            :py:meth:`write_data` will be automatically called when
            `sim.end()` is called.
        :param sampling_interval: a value in milliseconds, and an integer
            multiple of the simulation timestep.
        :param indexes: The indexes of neurons to record from.
            This is non-standard PyNN and equivalent to creating a view with
            these indexes and asking the View to record.
        """
        if variables is None:  # reset the list of things to record
            if sampling_interval is not None:
                raise ConfigurationException(
                    "Clash between parameters in record."
                    "variables=None turns off recording,"
                    "while sampling_interval!=None implies turn on recording")
            if indexes is not None:
                warn_once(
                    logger,
                    "View.record with variable None is non-standard PyNN. "
                    "Only the neurons in the view have their record turned "
                    "off. Other neurons already set to record will remain "
                    "set to record")

            # note that if record(None) is called, its a reset
            self.turn_off_all_recording(indexes)
            # handle one element vs many elements
        elif isinstance(variables, str):
            # handle special case of 'all'
            if variables == "all":
                self.__turn_on_all_record(sampling_interval, to_file, indexes)
            else:
                # record variable
                self.turn_on_record(
                    variables, sampling_interval, to_file, indexes)

        else:  # list of variables, so just iterate though them
            if "all" in variables:
                self.__turn_on_all_record(sampling_interval, to_file, indexes)
            else:
                for variable in variables:
                    self.turn_on_record(
                        variable, sampling_interval, to_file, indexes)

    def __turn_on_all_record(
            self, sampling_interval: Optional[int], to_file: IoDest,
            indexes: Optional[Collection[int]]) -> None:
        """
        :param sampling_interval: the interval to record them
        :param to_file: If set, a file to write to (by handle or name)
        :param indexes: List of indexes to record or `None` for all
        :raises SimulatorRunningException: If `sim.run` is currently running
        :raises SimulatorNotSetupException: If called before `sim.setup`
        :raises SimulatorShutdownException: If called after `sim.end`
        """
        warn_once(
            logger, 'record("all") is non-standard PyNN, and '
                    'therefore may not be portable to other simulators.')

        # iterate though all possible recordings for this vertex
        for variable in self.__vertex.get_recordable_variables():
            self.turn_on_record(
                variable, sampling_interval, to_file, indexes)

    def turn_on_record(
            self, variable: str, sampling_interval: Optional[int] = None,
            to_file: IoDest = None,
            indexes: Optional[Collection[int]] = None) -> None:
        """
        Tell the vertex to record data.

        :param variable: The variable to record, supported variables to
            record are: ``gsyn_exc``, ``gsyn_inh``, ``v``, ``spikes``.
        :param sampling_interval: the interval to record them
        :param to_file: If set, a file to write to (by handle or name)
        :param indexes: List of indexes to record or `None` for all
        :raises SimulatorRunningException: If `sim.run` is currently running
        :raises SimulatorNotSetupException: If called before `sim.setup`
        :raises SimulatorShutdownException: If called after `sim.end`
        """
        SpynnakerDataView.check_user_can_act()

        if variable not in self.__write_to_files_indicators:
            logger.warning("unrecognised recording variable")

        # update file writer
        self.__write_to_files_indicators[variable] = to_file

        if variable == "gsyn_exc":
            if not self.__vertex.conductance_based:
                warn_once(
                    logger, "You are trying to record the excitatory "
                    "conductance from a model which does not use conductance "
                    "input. You will receive current measurements instead.")
        elif variable == "gsyn_inh":
            if not self.__vertex.conductance_based:
                warn_once(
                    logger, "You are trying to record the inhibitory "
                    "conductance from a model which does not use conductance "
                    "input. You will receive current measurements instead.")

        # Tell the vertex to record
        self.__vertex.set_recording(variable, sampling_interval, indexes)

    @property
    def recording_label(self) -> str:
        """
        The label from the vertex is applicable or a default.
        """
        SpynnakerDataView.check_user_can_act()
        return self.__vertex.label or "!!UNLABELLED VERTEX!!"

    def turn_off_all_recording(
            self, indexes: Optional[Collection[int]] = None) -> None:
        """
        Turns off recording, is used by a pop saying ``.record()``.

        :param indexes:
        """
        for variable in self.__vertex.get_recordable_variables():
            self.__vertex.set_not_recording(variable, indexes)

    def extract_neo_block(
            self, variables: Names, view_indexes: Optional[Sequence[int]],
            clear: bool, annotations: Optional[Dict[str, Any]]) -> neo.Block:
        """
        Extracts block from the vertices and puts them into a Neo block.

        :param variables: the variables to extract
        :param view_indexes: the indexes to be included in the view
        :param clear: if the variables should be cleared after reading
        :param annotations:
            annotations to put on the Neo block
        :return: The Neo block
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the recording not setup correctly
        """
        SpynnakerDataView.check_user_can_act()

        block: Optional[neo.Block] = None
        for previous in range(SpynnakerDataView.get_reset_number()):
            block = self.__append_previous_segment(
                block, previous, variables, view_indexes, clear, annotations)

        # add to the segments the new block
        return self.__append_current_segment(
            block, variables, view_indexes, clear, annotations)

    def write_data(
            self, csv_file: str, variables: Optional[Names],
            view_indexes: Optional[Sequence[int]] = None,
            annotations: Optional[Dict[str, Any]] = None) -> None:
        """
        Extracts block from the vertices and puts them into a Neo block.

        :param variables: the variables to extract
        :param variables: the variables to extract
        :param view_indexes: the indexes to be included in the view
        :param annotations:
            annotations to put on the Neo block
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the recording not setup correctly
        """
        pop_label = self.__population.label

        wrote_metadata = False
        for segment in range(SpynnakerDataView.get_reset_number()):
            with NeoBufferDatabase.segement_db(segment) as db:
                if not wrote_metadata:
                    wrote_metadata = db.csv_block_metadata(
                        csv_file, pop_label, annotations)
                if wrote_metadata:
                    db.csv_segment(csv_file, pop_label, variables,
                                   view_indexes, allow_missing=True)

        if SpynnakerDataView.is_reset_last():
            if wrote_metadata:
                logger.warning(
                    "Due to the call directly after reset, "
                    "the data will only contain {} segments",
                    SpynnakerDataView.get_reset_number() - 1)
                return
            else:
                raise ConfigurationException(
                    f"Unable to write data for {pop_label}")

        with NeoBufferDatabase() as db:
            if not wrote_metadata:
                wrote_metadata = db.csv_block_metadata(
                    csv_file, pop_label, annotations)
            if wrote_metadata:
                db.csv_segment(csv_file, pop_label, variables,
                               view_indexes, allow_missing=False)
            else:
                raise ConfigurationException(
                    f"Unable to write data for {pop_label}")

    def __append_current_segment(
            self, block: neo.Block, variables: Names,
            view_indexes: Optional[Sequence[int]], clear: bool,
            annotations: Optional[Dict[str, Any]]) -> neo.Block:
        """
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the recording not setup correctly
        """
        with NeoBufferDatabase() as db:
            if block is None:
                block = db.get_empty_block(
                    self.__population.label, annotations)
                if block is None:
                    raise ConfigurationException(
                        f"No data for {self.__population.label}")
            if SpynnakerDataView.is_reset_last():
                logger.warning(
                    "Due to the call directly after reset, "
                    "the data will only contain {} segments",
                    SpynnakerDataView.get_reset_number() - 1)
            else:
                db.add_segment(
                    block, self.__population.label, variables, view_indexes,
                    allow_missing=False)
                if clear:
                    db.clear_data(self.__population.label, variables)
            return block

    def __append_previous_segment(
            self, block: Optional[neo.Block], segment_number: int,
            variables: Names, view_indexes: Optional[Sequence[int]],
            clear: bool,
            annotations: Optional[Dict[str, Any]]) -> Optional[neo.Block]:
        """
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the recording not setup correctly
        """
        with NeoBufferDatabase.segement_db(
                segment_number, read_only=not clear) as db:
            if block is None:
                block = db.get_empty_block(
                    self.__population.label, annotations)
            if block is not None:
                db.add_segment(
                    block, self.__population.label, variables, view_indexes,
                    allow_missing=True)
                if clear:
                    db.clear_data(self.__population.label, variables)
            return block



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from enum import Enum
from spynnaker.pyNN.external_devices_models\
    .abstract_multicast_controllable_device import (
        SendType)


class AbstractPushBotOutputDevice(Enum):
    """
    Superclass of all output device descriptors.
    """

    def __init__(
            self, value: int, protocol_property: property, min_value: int,
            max_value: Decimal, time_between_send: int,
            send_type: SendType = SendType.SEND_TYPE_INT):
        """
        :param value: Enum ID
        :param protocol_property: The protocol property of the output device
        :param min_value: smallest value allowed
        :param max_value: largest value allowed
        :param time_between_send: Time between sends
        :param send_type: The type of data to be sent.
        """
        self._value_ = value
        self._protocol_property = protocol_property
        self._min_value = min_value
        self._max_value = max_value
        self._time_between_send = time_between_send
        self._send_type = send_type

    @property
    def protocol_property(self) -> property:
        """
        The protocol property of the output device
        """
        return self._protocol_property

    @property
    def min_value(self) -> int:
        """
        The minimum value of the output device
        """
        return self._min_value

    @property
    def max_value(self) -> Decimal:
        """
        Max Value
        """
        return self._max_value

    @property
    def time_between_send(self) -> int:
        """
        Time between sends
        """
        return self._time_between_send

    @property
    def send_type(self) -> SendType:
        """
        The type of data to be sent.
        """
        return self._send_type



# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pacman.model.graphs import AbstractSupportsSDRAMEdges
from pacman.model.graphs.machine import (
    SourceSegmentedSDRAMMachinePartition)
from spinn_utilities.abstract_base import AbstractBase, abstractmethod


# pylint: disable=abstract-method
class SendsSynapticInputsOverSDRAM(
        AbstractSupportsSDRAMEdges, metaclass=AbstractBase):
    """
    An object that sends synaptic inputs over SDRAM.
    """

    @abstractmethod
    def set_sdram_partition(
            self,
            sdram_partition: SourceSegmentedSDRAMMachinePartition) -> None:
        """
        Set the SDRAM partition.  Must only be called once per instance.

        :param sdram_partition:
            The SDRAM partition to receive synapses from
        """
        raise NotImplementedError
